#!/usr/bin/env python3

import sys
import argparse

from AaronTools.geometry import Geometry
from AaronTools.fileIO import FileReader, read_types
from AaronTools.utils.utils import get_filename, glob_files

cube_parser = argparse.ArgumentParser(
    description="print a cube file for a molecular orbital",
    formatter_class=argparse.RawTextHelpFormatter
)

cube_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="an FCHK file, ORCA output with MO's, or NBO files"
)

cube_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    dest="outfile",
    help="output destination\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: stdout"
)

info = cube_parser.add_mutually_exclusive_group(required=False)
info.add_argument(
    "-mo", "--molecular-orbital",
    dest="mo_ndx",
    default="homo",
    help="index of molecular orbital to print (0-indexed)\n"
    "can also give 'homo' or 'lumo' for highest occupied or\n"
    "lowest unoccupied molecular orbital\n"
    "Default: highest occupied MO in the ground state"
)
info.add_argument(
    "-ao", "--atomic-orbital",
    dest="ao_ndx",
    default=None,
    help="index of atomic orbital to print (0-indexed)"
)

info.add_argument(
    "-ed", "--electron-density",
    dest="density",
    default=False,
    action="store_true",
    help="print electron density"
)

info.add_argument(
    "-fd", "--fukui-donor",
    dest="fukui_donor",
    default=False,
    action="store_true",
    help="print Fukui donor values\n"
    "see DOI 10.1002/jcc.24699 for weighting method"
)

info.add_argument(
    "-fa", "--fukui-acceptor",
    dest="fukui_acceptor",
    default=False,
    action="store_true",
    help="print Fukui acceptor values\n"
    "see DOI 10.1021/acs.jpca.9b07516 for weighting method"
)

info.add_argument(
    "-f2", "--fukui-dual",
    dest="fukui_dual",
    default=False,
    action="store_true",
    help="print Fukui dual values\n"
    "see DOI 10.1021/acs.jpca.9b07516 for weighting method"
)

cube_parser.add_argument(
    "-d", "--delta",
    type=float,
    dest="delta",
    default=0.1,
    help="delta parameter for weighting orbitals in Fukui functions\n"
    "Default: 0.1 Hartree",
)

cube_parser.add_argument(
    "-s", "--spacing",
    type=float,
    dest="spacing",
    default=0.2,
    help="spacing between points in the cube file\n"
    "Default: 0.2",
)

cube_parser.add_argument(
    "-p", "--padding",
    type=float,
    dest="padding",
    default=4,
    help="extra space around the molecule\n"
    "Default: 4"
)

cube_parser.add_argument(
    "-xyz", "--standard-axes",
    action="store_true",
    dest="xyz",
    default=False,
    help="use x, y, and z axes to define the directions\n"
    "Default: determine directions using SVD"
)

cube_parser.add_argument(
    "-nt", "--number-of-threads",
    type=int,
    default=1,
    dest="n_jobs",
    help="number of threads to use when evaluating basis functions"
    "this is on top of NumPy's multithreading,\n"
    "so if NumPy uses 8 threads and n_jobs=2, you can\n"
    "expect to see 16 threads in use\n"
    "Default: 1"
)

cube_parser.add_argument(
    "-nbo", "--nbo-file",
    type=str,
    default=None,
    dest="nbo_name",
    help="file containing coefficients for NBO's (e.g. *.37 file)"
    "ignored unless input file is a *.47 file"
)

cube_parser.add_argument(
    "-m", "--max-array",
    type=int,
    default=10000000,
    dest="max_length",
    help="max. array size to read from FCHK files\n"
    "a reasonable size for setting parsing orbital data\n"
    "can improve performance when reading large FCHK files\n"
    "too small of a value will prevent orbital data from\n"
    "being parsed\n"
    "Default: 10000000",
)


args = cube_parser.parse_args()

kind = args.mo_ndx

if args.density:
    kind = "density"
elif args.fukui_donor:
    kind = "fukui donor"
elif args.fukui_acceptor:
    kind = "fukui acceptor"
elif args.fukui_dual:
    kind = "fukui dual"
elif args.ao_ndx:
    kind = "AO %s" % args.ao_ndx
elif args.mo_ndx.isdigit():
    kind = "MO %s" % args.mo_ndx


for f in glob_files(args.infile, parser=cube_parser):
    if isinstance(f, str):
        infile = FileReader(
            f, just_geom=False, nbo_name=args.nbo_name, max_length=args.max_length
        )
    elif len(sys.argv) >= 1:
        infile = FileReader(
            ("from stdin", "fchk", f),
            just_geom=False,
            nbo_name=args.nbo_name,
            max_length=args.max_length,
        )

    geom = Geometry(infile, refresh_connected=False, refresh_ranks=False)

    out = geom.write(
        outfile=False,
        orbitals=infile.other["orbitals"],
        padding=args.padding,
        kind=kind,
        spacing=args.spacing,
        style="cube",
        xyz=args.xyz,
        delta=args.delta,
        n_jobs=args.n_jobs,
    )
    
    if not args.outfile:
        print(out)
    else:
        outfile = args.outfile
        if "$INFILE" in outfile:
            outfile = outfile.replace("$INFILE", get_filename(f))
        with open(outfile, "w") as f:
            f.write(out)




#!/usr/bin/env python3

import argparse
import sys

import numpy as np

from AaronTools.comp_output import CompOutput
from AaronTools.const import UNIT
from AaronTools.fileIO import FileReader
from AaronTools.geometry import Geometry
from AaronTools.substituent import Substituent
from AaronTools.utils.utils import boltzmann_coefficients, glob_files

def main(argv):
    sterimol_parser = argparse.ArgumentParser(
        description="calculate Boltzmann-weighted Sterimol parameters - see doi 10.1021/acscatal.8b04043",
        formatter_class=argparse.RawTextHelpFormatter
    )
    
    sterimol_parser.add_argument(
        "infiles", metavar="input files",
        type=str,
        nargs="+",
        help="file containing coordinates and energy"
    )
    
    sterimol_parser.add_argument(
        "-if", "--input-format",
        type=str,
        default=None,
        choices=["log", "out", "dat"],
        dest="input_format",
        help="file format of input"
    )
    
    sterimol_parser.add_argument(
        "-s", "--substituent-atom",
        type=str,
        required=True,
        dest="targets",
        help="substituent atom\n" +
        "1-indexed position of the starting position of the\n" +
        "substituent of which you are calculating sterimol\nparameters"
    )
    
    sterimol_parser.add_argument(
        "-a", "--attached-to",
        type=str,
        required=True,
        dest="avoid",
        help="non-substituent atom\n" +
        "1-indexed position of the starting position of the atom\n" +
        "connected to the substituent of which you are calculating\n" +
        "sterimol parameters"
    )
    
    sterimol_parser.add_argument(
        "-r", "--radii",
        type=str,
        default="bondi",
        choices=["bondi", "umn"],
        dest="radii",
        help="VDW radii to use in calculation\n" + 
        "umn: main group vdw radii from J. Phys. Chem. A 2009, 113, 19, 5806–5812\n" +
        "    (DOI: 10.1021/jp8111556)\n" + 
        "    transition metals are crystal radii from Batsanov, S.S. Van der Waals\n" +
        "    Radii of Elements. Inorganic Materials 37, 871–885 (2001).\n" +
        "    (DOI: 10.1023/A:1011625728803)\n" + 
        "bondi: radii from J. Phys. Chem. 1964, 68, 3, 441–451 (DOI: 10.1021/j100785a001)\n" +
        "Default: bondi"
    )
    
    sterimol_parser.add_argument(
        "-l", "--old-l",
        action="store_true",
        required=False,
        dest="old_L",
        help="approximate FORTRAN Sterimol method for determining L\n"
        "This is 0.4 + the ideal bond length for a target-H bond\n"
        "to outer VDW radii of atoms projected onto L-axis\n"
        "Default: L value is from VDW radii of target atom to outer\n"
        "VDW radii of atoms projected onto L-axis"
    )
    
    sterimol_parser.add_argument(
        "-t", "--temperature",
        type=float,
        default=298.15,
        required=False,
        dest="temperature",
        help="temperature in K\nDefault: 298.15"
    )
    
    sterimol_parser.add_argument(
        "-f", "--frequency",
        action="store_true",
        default=False,
        required=False,
        dest="frequency",
        help="input files are frequency job output files\n"
        "additional average values will be calculated for ZPE, H, G, etc."
    )
    
    sterimol_parser.add_argument(
        "-w0",
        "--frequency-cutoff",
        type=float,
        default=100.0,
        required=False,
        dest="w0",
        help="cutoff frequency for quasi free energy corrections (1/cm)\n" +
        "Default: 100 cm^-1"
    )
    
    sterimol_parser.add_argument(
        "-v", "--verbose",
        action="store_true",
        default=False,
        required=False,
        dest="verbose",
        help="also print population"
    )
    
    sterimol_parser.add_argument(
        "-o", "--output",
        type=str,
        default=False,
        required=False,
        metavar="output destination",
        dest="outfile",
        help="output destination\n" +
        "Default: stdout"
    )
    
    
    args = sterimol_parser.parse_args(args=argv)
    
    subs = []
    energies = {"E":[]}
    if args.frequency:
        energies["E+ZPE"] = []
        energies["H(RRHO)"] = []
        energies["G(RRHO)"] = []
        energies["G(Quasi-RRHO)"] = []
        energies["G(Quasi-Harmonic)"] = []
    
    for infile in glob_files(args.infiles, parser=sterimol_parser):
        if args.input_format is not None:
            fr = FileReader((infile, args.input_format, infile), just_geom=False)
        else:
            fr = FileReader(infile, just_geom=False)
    
        geom = Geometry(fr)
        target = args.targets
        avoid = args.avoid
        end = geom.find(avoid)[0]
        frag = geom.get_fragment(target, stop=end)
        sub = Substituent(frag, end=end, detect=False)
        subs.append(sub)
        
        nrg = fr.other["energy"]
        energies["E"].append(nrg)
        if args.frequency:
            co = CompOutput(fr)
            dE, dH, entropy = co.therm_corr(temperature=args.temperature)
            rrho_dG = co.calc_G_corr(v0=0, temperature=args.temperature, method="RRHO")
            qrrho_dG = co.calc_G_corr(v0=args.w0, temperature=args.temperature, method="QRRHO")
            qharm_dG = co.calc_G_corr(v0=args.w0, temperature=args.temperature, method="QHARM")
            energies["E+ZPE"].append(nrg + co.ZPVE)
            energies["H(RRHO)"].append(nrg + dH)
            energies["G(RRHO)"].append(nrg + rrho_dG)
            energies["G(Quasi-RRHO)"].append(nrg + qrrho_dG)
            energies["G(Quasi-Harmonic)"].append(nrg + qharm_dG)
    
    s = ""
    for nrg_type in energies:
        energies_arr = np.array(energies[nrg_type])
        energies_arr *= UNIT.HART_TO_KCAL

        if args.verbose and nrg_type == "E":
            s += "\t".join(["B1", "B2", "B3", "B4", "B5", "L", "file"])
            s += "\n"
            for f, sub in zip(args.infiles, subs):
                data = sub.sterimol(
                    radii=args.radii,
                    old_L=args.old_L,
                )
                s += "\t".join(
                    ["%.2f" % data[x] for x in ["B1", "B2", "B3", "B4", "B5", "L"]]
                )
                s += "\t%s\n" % f
        s += "weighted using %s:\n" % nrg_type
        data = Substituent.weighted_sterimol(
            subs,
            energies_arr,
            args.temperature,
            radii=args.radii,
            old_L=args.old_L,
        )
        if args.verbose:
            coeff = boltzmann_coefficients(energies_arr, args.temperature)
            coeff /= sum(coeff)
            coeff *= 100
            for f, c, e in zip(args.infiles, coeff, energies_arr):
                s += "%s  %.1f%% (%.1f kcal/mol)\n" % (f, c, e - min(energies_arr))
                
        s += "\t".join(["B1", "B2", "B3", "B4", "B5", "L"])
        s += "\n"
        s += "\t".join(["%.2f" % data[x] for x in ["B1", "B2", "B3", "B4", "B5", "L"]])
        s += "\n"
        s += "\n"
    
    if not args.outfile:
        print(s)
    else:
        with open(args.outfile, "w") as f:
            f.write(s)

if __name__ == "__main__":
    main(None)



#!/usr/bin/env python3

import sys
import argparse

from AaronTools.geometry import Geometry
from AaronTools.component import Component
from AaronTools.finders import AnyTransitionMetal
from AaronTools.fileIO import FileReader, read_types
from AaronTools.utils.utils import glob_files

cone_parser = argparse.ArgumentParser(
    description="calculate ligand cone angles",
    formatter_class=argparse.RawTextHelpFormatter
)

cone_parser = argparse.ArgumentParser(
    description="print ligand cone angle",
    formatter_class=argparse.RawTextHelpFormatter
)

cone_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file",
)

cone_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    required=False,
    dest="outfile",
    help="output destination\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: stdout"
)

cone_parser.add_argument(
    "-if", "--input-format",
    type=str,
    default=None,
    dest="input_format",
    choices=read_types,
    help="file format of input - xyz is assumed if input is stdin",
)

cone_parser.add_argument(
    "-k", "--key-atoms",
    type=str,
    default=None,
    dest="key_atoms",
    help="indices of ligand coordinating atoms you are calculating\n" +
    "the cone angle of (1-indexed)",
)

cone_parser.add_argument(
    "-c", "--center",
    type=str,
    default=AnyTransitionMetal(),
    dest="center",
    help="index of complex's center atom (1-indexed)\nDefault: transition metals",
)

cone_parser.add_argument(
    "-m", "--method",
    type=lambda x: x.capitalize(),
    choices=["Tolman", "Exact"],
    default="exact",
    dest="method",
    help="cone angle type\n" +
    "Tolman: Tolman's method for unsymmetric mono- and bidentate ligands\n" +
    "        see J. Am. Chem. Soc. 1974, 96, 1, 53–60 (DOI:\n" +
    "        10.1021/ja00808a009)\n" +
    "Exact: (Default) Allen's method for an all-encompassing cone\n" +
    "       see Bilbrey, J.A., Kazez, A.H., Locklin, J. and Allen, W.D.\n" +
    "       (2013), Exact ligand cone angles. J. Comput. Chem., 34:\n" +
    "       1189-1197. (DOI: 10.1002/jcc.23217)",
)

cone_parser.add_argument(
    "-r", "--vdw-radii",
    default="umn",
    choices=["umn", "bondi"],
    dest="radii",
    help="VDW radii to use in calculation\n" + 
    "umn: main group vdw radii from J. Phys. Chem. A 2009, 113, 19, 5806–5812\n" +
    "    (DOI: 10.1021/jp8111556)\n" + 
    "    transition metals are crystal radii from Batsanov, S.S. Van der Waals\n" +
    "    Radii of Elements. Inorganic Materials 37, 871–885 (2001).\n" +
    "    (DOI: 10.1023/A:1011625728803)\n" + 
    "bondi: radii from J. Phys. Chem. 1964, 68, 3, 441–451 (DOI: 10.1021/j100785a001)\n" +
    "Default: umn",
)

cone_parser.add_argument(
    "-b", "--cone-bild",
    action="store_true",
    default=False,
    dest="print_cones",
    help="print Chimera/ChimeraX bild file containing cones",
)

args = cone_parser.parse_args()

s = ""

for f in glob_files(args.infile, parser=cone_parser):
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format[0], None))
        else:
            infile = FileReader(f, just_geom=False)
    else:
        if args.input_format is not None:
            infile = FileReader(("from stdin", args.input_format[0], f))
        else:
            if len(sys.argv) >= 1:
                infile = FileReader(("from stdin", "xyz", f))

    geom = Geometry(infile)

    ligand = geom.get_fragment(args.key_atoms, stop=args.center)

    comp = Component(ligand, key_atoms=args.key_atoms, detect_backbone=False)

    angle = comp.cone_angle(
        center=geom.find(args.center),
        method=args.method,
        radii=args.radii,
        return_cones=args.print_cones,
    )

    if args.print_cones:
        angle, cones = angle

    if len(args.infile) > 1:
        s += "%20s:\t" % f

    s += "%4.1f\n" % angle
    
    if args.print_cones:
        s += ".transparency 0.5\n"
        for cone in cones:
            apex, base, radius = cone
            s += ".cone   %6.3f %6.3f %6.3f   %6.3f %6.3f %6.3f   %.3f open\n" % (
                *apex, *base, radius
            )

if not args.outfile:
    print(s.rstrip())
else:
    with open(args.outfile, "a") as f:
        f.write(s.rstrip())



#!/usr/bin/env python3

import sys
import argparse

from AaronTools.geometry import Geometry
from AaronTools.fileIO import FileReader, read_types
from AaronTools.finders import *
from AaronTools.utils.utils import glob_files

vsepr_choices = [
    "linear_1",
    "linear_2",
    "bent_2_planar",
    "bent_2_tetrahedral",
    "trigonal_planar",
    "bent_3_tetrahedral",
    "t_shaped",
    "tetrahedral",
    "sawhorse",
    "square_planar",
    "trigonal_bipyriamidal",
    "square_pyramidal",
    "octahedral",
]

find_parser = argparse.ArgumentParser(
    description="find atoms matching a description and return the list of indices",
    formatter_class=argparse.RawTextHelpFormatter
)

find_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file"
)

find_parser.add_argument(
    "-if", "--input-format",
    type=str,
    default=None,
    dest="input_format",
    choices=read_types,
    help="file format of input - xyz is assumed if input is stdin",
)


find_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    required=False,
    dest="outfile",
    help="output destination\n" +
    "Default: stdout"
)

find_parser.add_argument(
    "-d",
    "--delimiter",
    required=False,
    dest="delim",
    default="comma",
    choices=["comma", "semicolon", "tab", "space"],
    help="delimiter for output atom indices"
)

find_parser.add_argument(
    "-e", "--element",
    type=str,
    action="append",
    default=[],
    required=False,
    dest="elements",
    help="element symbol",
)

find_parser.add_argument(
    "-n", "--index",
    type=str,
    action="append",
    default=[],
    required=False,
    dest="ndx",
    help="1-index position of atoms in the input file\n"
    "may hyphen separated to denote a range\n"
    "ranges and individual indices may be comma-separated",
)

find_parser.add_argument(
    "-bf", "--bonds-from",
    type=str,
    nargs=2,
    action="append",
    default=[],
    required=False,
    metavar=("BONDS", "NUM"),
    dest="bonds_from",
    help="find atoms BONDS (integer) bonds away from atom NDX",
)

find_parser.add_argument(
    "-wb", "--within-bonds",
    type=str,
    nargs=2,
    action="append",
    default=[],
    required=False,
    metavar=("BONDS", "NDX"),
    dest="within_bonds",
    help="find atoms within BONDS (integer) bonds from atom NDX",
)

find_parser.add_argument(
    "-bt", "--bonded-to",
    type=str,
    action="append",
    default=[],
    required=False,
    metavar="NDX",
    dest="bonded_to",
    help="find atoms bonded to atom NDX",
)

find_parser.add_argument(
    "-pd", "--point-distance",
    type=float,
    nargs=4,
    action="append",
    default=[],
    required=False,
    metavar=("X", "Y", "Z", "R"),
    dest="point_dist",
    help="find atoms within R Angstroms of (X, Y, Z)",
)

find_parser.add_argument(
    "-ad", "--atom-distance",
    type=str,
    nargs=2,
    action="append",
    default=[],
    required=False,
    metavar=("NDX", "R"),
    dest="atom_dist",
    help="find atoms within R Angstroms of atom NDX",
)

find_parser.add_argument(
    "-tm", "--transition-metal",
    action="store_true",
    default=False,
    required=False,
    dest="tmetal",
    help="find any elements in the d-block, up to the Actinides",
)

find_parser.add_argument(
    "-mg", "--main-group",
    action="store_true",
    default=False,
    required=False,
    dest="main_group",
    help="find any main group element (including H)",
)

find_parser.add_argument(
    "-v", "--vsepr",
    type=str,
    action="append",
    default=[],
    required=False,
    choices=vsepr_choices,
    metavar="SHAPE",
    dest="vsepr",
    help="find atoms with the specified VSEPR shape\n" +
    "shape can be:\n%s" % "".join(
        s + ", " if (sum(len(x) for x in vsepr_choices[:i])) % 40 < 21 else s + ",\n"
        for i, s in enumerate(vsepr_choices)
    ).strip().strip(","),
)

find_parser.add_argument(
    "-nb", "--number-of-bonds",
    type=int,
    action="append",
    default=[],
    required=False,
    dest="num_neighbors",
    help="find atoms with the specified number of bonds",
)

find_parser.add_argument(
    "-ct", "--closer-to",
    nargs=2,
    metavar=("THIS_NDX", "THAN_NDX"),
    action="append",
    default=[],
    dest="closer_to",
    help="atoms that are fewer bonds from THIS_NDX than THAN_NDX",
)

find_parser.add_argument(
    "-c", "--chiral-center",
    action="store_true",
    default=False,
    dest="chiral",
    help="find chiral centers",
)

find_parser.add_argument(
    "-f", "--fragment",
    type=str,
    default=None,
    dest="fragments",
    help="fragments containing the specified atoms"
)

finder_combination = find_parser.add_argument_group(
    "match method (Default is atoms matching all)"
)
finder_combination.add_argument(
    "-or", "--match-any",
    action="store_true",
    default=False,
    dest="match_any",
    help="find atoms matching any of the given descriptions",
)

finder_combination.add_argument(
    "-i", "--invert",
    action="store_true",
    default=False,
    dest="invert",
    help="invert match results",
)


args = find_parser.parse_args()

# create a list of geometry-independent finders
finders = []
for ndx in args.ndx:
    finders.append(ndx)

for ele in args.elements:
    finders.append(ele)

for pd in args.point_dist:
    r = pd.pop(-1)
    p = np.array(pd)
    finders.append(WithinRadiusFromPoint(pd, r))

if args.tmetal:
    finders.append(AnyTransitionMetal())

if args.main_group:
    finders.append(AnyNonTransitionMetal())

for vsepr in args.vsepr:
    finders.append(VSEPR(vsepr.replace("_", " ")))

for nb in args.num_neighbors:
    finders.append(NumberOfBonds(nb))

if args.chiral:
    finders.append(ChiralCenters())

if args.delim == "comma":
    delim = ","
elif args.delim == "space":
    delim = " "
elif args.delim == "tab":
    delim = "\t"
elif args.delim == "semicolon":
    delim = ";"

s = ""

for f in glob_files(args.infile, parser=find_parser):
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format, None))
        else:
            infile = FileReader(f)
    else:
        if args.input_format is not None:
            infile = FileReader(("from stdin", args.input_format, f))
        else:
            if len(sys.argv) >= 1:
                infile = FileReader(("from stdin", "xyz", f))

    geom = Geometry(infile)

    geom_finders = [x for x in finders]
    # some finders require and atom upon instantiation
    # add those once we have the geometry
    for ad in args.atom_dist:
        r = float(ad.pop(-1))
        atom_ndx = ad.pop(0)
        for atom in geom.find(atom_ndx):
            geom_finders.append(WithinRadiusFromAtom(atom, r))

    for wb in args.within_bonds:
        n_bonds = int(wb.pop(0))
        atom_ndx = wb.pop(0)
        for atom in geom.find(atom_ndx):
            geom_finders.append(WithinBondsOf(atom, n_bonds))

    for bt in args.bonded_to:
        for atom in geom.find(bt):
            geom_finders.append(BondedTo(atom))
    
    for ct in args.closer_to:
        atom1, atom2 = geom.find(ct)
        geom_finders.append(CloserTo(atom1, atom2))
    
    if args.fragments:
        for atom in geom.find(args.fragments):
            frag_atoms = geom.get_all_connected(atom)
            geom_finders.append(frag_atoms)

    if len(args.infile) > 1:
        s += "%s\n" % str(s)

    try:
        if args.match_any:
            results = geom.find(geom_finders)
        else:
            results = geom.find(*geom_finders)

        if args.invert:
            results = geom.find(NotAny(results))

        s += delim.join(atom.name for atom in results)
        s += "\n"
    except LookupError as e:
        if args.invert:
            s += delim.join(atom.name for atom in geom.atoms)
            s += "\n"
        else:
            s += "%s\n" % str(e)


if not args.outfile:
    print(s.strip())
else:
    with open(args.outfile, "a") as f:
        f.write(s.strip())



"""used for specifying emperical dispersion for Theory() instances"""

from AaronTools import addlogger
from AaronTools.theory import GAUSSIAN_ROUTE, ORCA_ROUTE, QCHEM_REM

@addlogger
class EmpiricalDispersion:
    """try to keep emerpical dispersion keywords and settings consistent across file types"""
    
    LOG = None
    
    def __init__(self, name):
        """name can be (availability may vary):
            Grimme D2 (or D2, -D2, GD2)
            Zero-damped Grimme D3 (or D3, -D3, GD3)
            Becke-Johnson damped Grimme D3 (or D3BJ, -D3BJ, GD3BJ)
            Becke-Johnson damped modified Grimme D3 (or B3MBJ, -D3MBJ)
            Petersson-Frisch (or PFD)
            Grimme D4 (or D4, -D4, GD4)
            Chai & Head-Gordon (or CHG, -CHG)
            Nonlocal Approximation (or NL, NLA, -NL)
            Pernal, Podeszwa, Patkowski, & Szalewicz (or DAS2009, -DAS2009)
            Podeszwa, Katarzyna, Patkowski, & Szalewicz (or DAS2010, -DAS2010)
            Coupled-Cluster Doubles (or CCD)
            Řezác, Greenwell, & Beran (or DMP2)
            Coupled-Cluster Doubles + Řezác, Greenwell, & Beran (or (CCD)DMP2)

        or simply the keyword for the input file type you are using"""

        self.name = name

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False
        if self.name == other.name:
            return True
        for d in [
                ["grimme d2", "d2", "-d2", "gd2"],
                ["grimme d3", "d3", "-d3", "gd3", "d3zero", "zero-damped grimme d3"],
                ["becke-johnson damped grimme d3", "d3bj", "-d3bj", "gd3bj"],
                ["becke-johnson damped modified grimme d3", "d3mbj", "-d3mbj"],
                ["petersson-frisch", "pfd"],
                ["grimme d4", "d4", "-d4", "gd4"],
                ["nonlocal approximation", "nl", "nla", "-nl"],
                ["coupled-cluster doubles", "ccd"],
        ]:
            if self.name.lower() in d and other.name.lower() in d:
                return True
        
        return False

    def copy(self):
        new_dict = dict()
        for key, value in self.__dict__.items():
            try:
                new_dict[key] = value.copy()
            except AttributeError:
                new_dict[key] = value
                # ignore chimerax objects so seqcrow doesn't print a
                # warning when a geometry is copied
                if "chimerax" in value.__class__.__module__:
                    continue
                if value.__class__.__module__ != "builtins":
                    self.LOG.warning(
                        "No copy method for {}: in-place changes may occur".format(
                            type(value)
                        )
                    )
        
        return self.__class__(**new_dict)

    def get_gaussian(self):
        """Acceptable dispersion methods for Gaussian are:
        Grimme D2
        Grimme D3
        Becke-Johnson damped Grimme D3
        Petersson-Frisch

        Dispersion methods available in other software that will be modified are:
        Grimme D4
        undampened Grimme D3"""

        if any(
                self.name.upper() == name for name in [
                    "GRIMME D2", "GD2", "D2", "-D2"
                ]
        ):
            return ({GAUSSIAN_ROUTE:{"EmpiricalDispersion":["GD2"]}}, None)
        elif any(
                self.name.upper() == name for name in [
                    "ZERO-DAMPED GRIMME D3", "GRIMME D3", "GD3", "D3", "-D3", "D3ZERO"
                ]
        ):
            return ({GAUSSIAN_ROUTE:{"EmpiricalDispersion":["GD3"]}}, None)
        elif any(
                self.name.upper() == name for name in [
                    "BECKE-JOHNSON DAMPED GRIMME D3", "GD3BJ", "D3BJ", "-D3BJ"
                ]
        ):
            return ({GAUSSIAN_ROUTE:{"EmpiricalDispersion":["GD3BJ"]}}, None)
        elif any(
                self.name.upper() == name for name in [
                    "PETERSSON-FRISCH", "PFD"
                ]
        ):
            return ({GAUSSIAN_ROUTE:{"EmpiricalDispersion":["PFD"]}}, None)

        #dispersions in ORCA but not Gaussian
        elif self.name == "Grimme D4":
            out_dict = {
                GAUSSIAN_ROUTE: {
                    "EmpiricalDispersion":["GD3BJ"]
                }
            }
            return (out_dict, "Grimme's D4 has no keyword in Gaussian, switching to GD3BJ")

        #unrecognized
        return (self.name, "unrecognized emperical dispersion: %s" % self.name)

    def get_orca(self):
        """Acceptable keywords for ORCA are:
        Grimme D2
        Zero-damped Grimme D3
        Becke-Johnson damped Grimme D3
        Grimme D4"""
        if any(
                self.name.upper() == name for name in [
                    "GRIMME D2", "GD2", "D2", "-D2"
                ]
        ):
            return ({ORCA_ROUTE:["D2"]}, None)
        elif any(
                self.name.upper() == name for name in [
                    "ZERO-DAMPED GRIMME D3", "GRIMME D3", "GD3", "D3", "-D3", "D3ZERO"
                ]
        ):
            return ({ORCA_ROUTE:["D3ZERO"]}, None)
        elif any(
                self.name.upper() == name for name in [
                    "BECKE-JOHNSON DAMPED GRIMME D3", "GD3BJ", "D3BJ", "-D3BJ"
                ]
        ):
            return ({ORCA_ROUTE:["D3BJ"]}, None)
        elif any(
                self.name.upper() == name for name in [
                    "GRIMME D4", "GD4", "D4", "-D4"
                ]
        ):
            return ({ORCA_ROUTE:["D4"]}, None)

        out_dict = {
            ORCA_ROUTE: [self.name]
        }
        return(out_dict, "unrecognized emperical dispersion: %s" % self.name)

    def get_psi4(self):
        """Acceptable keywords for Psi4 are:
        Grimme D1
        Grimme D2
        Zero-damped Grimme D3
        Becke-Johnson damped Grimme D3
        Chai & Head-Gordon
        Nonlocal Approximation
        Pernal, Podeszwa, Patkowski, & Szalewicz
        Podeszwa, Katarzyna, Patkowski, & Szalewicz
        Řezác, Greenwell, & Beran"""
        if any(
                self.name.upper() == name for name in [
                    "GRIMME D1", "GD1", "D1", "-D1"
                ]
        ):
            return ("-d1", None)
        elif any(
                self.name.upper() == name for name in [
                    "GRIMME D2", "GD2", "D2", "-D2"
                ]
        ):
            return ("-d2", None)
        elif any(
                self.name.upper() == name for name in [
                    "ZERO-DAMPED GRIMME D3", "GRIMME D3", "GD3", "D3", "-D3", "D3ZERO"
                ]
        ):
            return ("-d3", None)
        elif any(
                self.name.upper() == name for name in [
                    "BECKE-JOHNSON DAMPED GRIMME D3", "GD3BJ", "D3BJ", "-D3BJ"
                ]
        ):
            return ("-d3bj", None)
        elif any(
                self.name.upper() == name for name in [
                    "BECKE-JOHNSON DAMPED MODIFIED GRIMME D3", "GD3MBJ", "D3MBJ", "-D3MBJ"
                ]
        ):
            return ("-d3mbj", None)
        elif any(
                self.name.upper() == name for name in [
                    "CHAI & HEAD-GORDON", "CHG", "-CHG"
                ]
        ):
            return ("-chg", None)
        elif any(
                self.name.upper() == name for name in [
                    "NONLOCAL APPROXIMATION", "NL", "NLA", "-NL"
                ]
        ):
            return ("-nl", None)
        elif any(
                self.name.upper() == name for name in [
                    "PERNAL, PODESZWA, PATKOWSKI, & SZALEWICZ", "DAS2009", "-DAS2009"
                ]
        ):
            return ("-das2009", None)
        elif any(
                self.name.upper() == name for name in [
                    "PODESZWA, KATARZYNA, PATKOWSKI, & SZALEWICZ", "DAS2010", "-DAS2010"
                ]
        ):
            return ("-das2010", None)
        elif any(
                self.name.upper() == name for name in [
                    "COUPLED-CLUSTER DOUBLES", "CCD"
                ]
        ):
            return ("(ccd)", None)
        elif any(
                self.name.upper() == name for name in [
                    "ŘEZÁC, GREENWELL, & BERAN", "DMP2"
                ]
        ):
            return ("dmp2", None)
        elif any(
                self.name.upper() == name for name in [
                    "COUPLED-CLUSTER DOUBLES + ŘEZÁC, GREENWELL, & BERAN", "(CCD)DMP2"
                ]
        ):
            return ("(ccd)dmp2", None)
        else:
            return (self.name, "unrecognized emperical dispersion: %s" % self.name)

    def get_qchem(self):
        """Acceptable keywords for QChem are:
        Grimme D2
        Modified Zero-damped Grimme D3
        Zero-damped Grimme D3
        Becke-Johnson damped Grimme D3
        Becke-Johnson damped modified Grimme D3
        Chai & Head-Gordon
        """
        if any(
                self.name.upper() == name for name in [
                    "GRIMME D2", "GD2", "D2", "-D2", "EMPIRICAL_GRIMME",
                ]
        ):
            return ({QCHEM_REM: {"DFT_D": "EMPIRICAL_GRIMME"}}, None)
        elif any(
                self.name.upper() == name for name in [
                    "ZERO-DAMPED GRIMME D3", "GRIMME D3", "GD3", "D3",
                    "-D3", "D3ZERO", "D3_ZERO",
                ]
        ):
            return ({QCHEM_REM: {"DFT_D": "D3_ZERO"}}, None)
        elif any(
                self.name.upper() == name for name in [
                    "MODIFIED ZERO-DAMPED GRIMME D3", "D3_ZEROM",
                ]
        ):
            # Smith et. al. modified D3
            return ({QCHEM_REM: {"DFT_D": "D3_ZEROM"}}, None)
        elif any(
                self.name.upper() == name for name in [
                    "C6 ONLY GRIMME D3", "C<SUB>6</SUB> ONLY GRIMME D3", "D3_CSO",
                ]
        ):
            return ({QCHEM_REM: {"DFT_D": "D3_CSO"}}, None)
        elif any(
                self.name.upper() == name for name in [
                    "OPTIMIZED POWER GRIMME D3", "D3_OP",
                ]
        ):
            return ({QCHEM_REM: {"DFT_D": "D3_OP"}}, None)
        elif any(
                self.name.upper() == name for name in [
                    "BECKE-JOHNSON DAMPED GRIMME D3", "GD3BJ", "D3BJ",
                    "-D3BJ", "B3BJ",
                ]
        ):
            return ({QCHEM_REM: {"DFT_D": "D3_BJ"}}, None)
        elif any(
                self.name.upper() == name for name in [
                    "BECKE-JOHNSON DAMPED MODIFIED GRIMME D3", "GD3MBJ",
                    "D3MBJ", "-D3MBJ", "D3_BJM",
                ]
        ):
            return ({QCHEM_REM: {"DFT_D": "D3_BJM"}}, None)
        elif any(
                self.name.upper() == name for name in [
                    "CHAI & HEAD-GORDON", "CHG", "-CHG", "EMPIRICAL_CHG"
                ]
        ):
            return ({QCHEM_REM: {"DFT_D": "EMPIRICAL_CHG"}}, None)
        elif any(
                self.name.upper() == name for name in [
                    "CALDEWEYHER ET. AL. D4", "D4",
                ]
        ):
            return ({QCHEM_REM: {"DFT_D": "D4"}}, None)
        else:
            return (self.name, "unrecognized emperical dispersion: %s" % self.name)



#!/usr/bin/env python3
import os
import unittest

from AaronTools.geometry import Geometry
from AaronTools.pathway import Pathway
from AaronTools.test import TestWithTimer, prefix, rmsd_tol
from numpy import array


class TestPathway(TestWithTimer):
    t60 = Geometry(os.path.join(prefix, "test_files", "torsion-60.xyz"))
    t90 = Geometry(os.path.join(prefix, "test_files", "torsion-90.xyz"))

    def test_interpolating_structure(self):
        # test to see if interpolated geometry is correct
        ref = Geometry(
            os.path.join(prefix, "ref_files", "torsion_interpolation.xyz")
        )
        pathway = Pathway(self.t60, array([self.t60.coords, self.t90.coords]))
        geom = pathway.geom_func(0.4)
        rmsd = geom.RMSD(ref, align=True, sort=False)
        self.assertTrue(rmsd < rmsd_tol(ref, superLoose=True))


def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestPathway("test_interpolating_structure"))
    return suite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



"""For parsing and storing atom information"""
import json
import os
import re

import numpy as np

from AaronTools import addlogger
from AaronTools.const import (
    CONNECTIVITY,
    EIJ,
    ELEMENTS,
    MASS,
    RADII,
    RIJ,
    SATURATION,
    TMETAL,
    VDW_RADII,
)

warn_LJ = set([])


class BondOrder:
    bonds = {}
    warn_atoms = set([])
    warn_str = (
        "\n"
        + "  Could not get bond order for: "
        + "    {} {}"
        + "    using bond order of 1"
    )
    # warn(s.format(a1.element, a2.element))

    def __init__(self):
        if BondOrder.bonds:
            return
        with open(
            os.path.join(
                os.path.dirname(__file__), "calculated_bond_lengths.json"
            )
        ) as f:
            BondOrder.bonds = json.load(f)

    @classmethod
    def key(cls, a1, a2):
        if isinstance(a1, Atom):
            a1 = a1.element
        if isinstance(a2, Atom):
            a2 = a2.element

        return " ".join(sorted([a1, a2]))

    @classmethod
    def get(cls, a1, a2):
        """determines bond order between two atoms based on bond length"""
        try:
            bonds = cls.bonds[cls.key(a1, a2)]
        except KeyError:
            if a1.element == "H" or a2.element == "H":
                return 1
            else:
                BondOrder.warn_atoms.add((a1.element, a2.element))
                return 1
        dist = a1.dist(a2)
        closest = 0, None  # (bond order, length diff)
        for order, length in bonds.items():
            diff = abs(length - dist)
            if closest[1] is None or diff < closest[1]:
                closest = order, diff
        return float(closest[0])


@addlogger
class Atom:
    """
    Attributes:
        element         str
        coords          np.array(float)
        flag            bool            true if frozen, false if relaxed
        name            str             form of \d+(\.\d+)*
        tags            set
        charge          float
        connected       set(Atom)
        constraint      set(Atom)       for determining constrained bonds
        _rank
        _radii          float           for calculating if bonded
        _connectivity   int             max connections without hypervalence
        _saturation     int             max connections without hypervalence or charges
    """

    LOG = None

    _bo = BondOrder()

    def __init__(
        self, element="", coords=None, flag=False, name="", tags=None, charge=None
    ):
        super().__setattr__("_hashed", False)
        if coords is None:
            coords = []
        if tags is None:
            tags = []
        # for BqO to have a ghost atom with oxygen basis functions
        ele = str(element).strip()
        element = ele.capitalize()
        if "-" in ele:
            element = "-".join(e.capitalize() for e in ele.split("-"))
        if element == "":
            self.element = element
            self._radii = None
            self._connectivity = None
        elif element not in ELEMENTS and not element.endswith("Bq"):
            raise ValueError("Unknown element detected: %s" % element)
        else:
            self.element = element
            self.reset()

        self.coords = np.array(coords, dtype=float)
        self.flag = bool(flag)
        self.name = str(name).strip()

        if hasattr(tags, "__iter__") and not isinstance(tags, str):
            self.tags = set(tags)
        else:
            self.tags = set([tags])

        self.charge = charge
        if charge:
            self.charge = float(charge)

        self.connected = set([])
        self.constraint = set([])
        self._rank = None

    # utilities
    def __float__(self):
        """
        converts self.name from a string to a floating point number
        """
        rv = self.name.split(".")
        if len(rv) == 0:
            return float(0)
        if len(rv) == 1:
            return float(rv[0])
        rv = "{}.{}".format(rv[0], rv[1])
        return float(rv)

    def __lt__(self, other):
        """
        sorts by canonical smiles rank, then by neighbor ID, then by name
            more connections first
            then, more non-H bonds first
            then, higher atomic number first
            then, higher number of attached hydrogens first
            then, lower sorting name first
        """
        if (
            self._rank is not None
            and other._rank is not None
            and self._rank != other._rank
        ):
            return self._rank > other._rank

        if self._rank is None or other._rank is None:
            # if the ranks are the same, we have little reason to
            # believe the invariants will differ
            # print("getting invariants during <", self._rank, other._rank)
            a = self.get_invariant()
            b = other.get_invariant()
            if a != b:
                return a > b

        # print("using names")
        a = self.name.split(".")
        b = other.name.split(".")
        while len(a) < len(b):
            a += ["0"]
        while len(b) < len(a):
            b += ["0"]
        for i, j in zip(a, b):
            try:
                if int(i) != int(j):
                    return int(i) < int(j)
            except ValueError:
                pass
        return True

    def __str__(self):
        s = ""
        s += "{:>4s} ".format(self.name)
        s += "{:<3s} ".format(self.element)
        for c in self.coords:
            s += " {: 10.6f}".format(c)
        return s

    def __repr__(self):
        s = ""
        s += "{:>3s} ".format(self.element)
        for c in self.coords:
            s += " {: 13.8f}".format(c)
        s += "  {: 2d}".format(-1 if self.flag else 0)
        s += " {:>4s}".format(self.name)
        s += " ({:d})".format(self._rank) if self._rank is not None else ""
        return s

    def _set_radii(self):
        """Sets atomic radii"""
        try:
            self._radii = float(RADII[self.element])
        except KeyError:
            self.LOG.warning("Radii not found for element: %s" % self.element)
        return

    def __setattr__(self, attr, val):
        if (
            not self._hashed
            or (attr == "_hashed" and val)
            or (attr != "element" and attr != "coords")
        ):
            super().__setattr__(attr, val)
        else:
            raise RuntimeError(
                "Atom %s's Geometry has been hashed and can no longer be changed\n"
                % self.name
                + "setattr was called to set %s to %s" % (attr, val)
            )

    def _set_vdw(self):
        """Sets atomic radii"""
        try:
            self._vdw = float(VDW_RADII[self.element])
        except KeyError:
            self.LOG.warning(
                "VDW Radii not found for element: %s" % self.element
            )
            self._vdw = 0
        return

    def _set_connectivity(self):
        """Sets theoretical maximum connectivity.
        If # connections > self._connectivity, then atom is hyper-valent
        """
        try:
            self._connectivity = int(CONNECTIVITY[self.element])
        except KeyError:
            pass
            # self.LOG.warning(
            #     "Connectivity not found for element: " + self.element
            # )
        return

    def _set_saturation(self):
        """Sets theoretical maximum connectivity without the atom having a formal charge.
        If # connections > self._saturation, then atom is hyper-valent or has a non-zero formal charge
        """
        try:
            self._saturation = int(SATURATION[self.element])
        except KeyError:
            if self.element not in TMETAL:
                self.LOG.warning(
                    "Saturation not found for element: " + self.element
                )
        return

    @property
    def is_dummy(self):
        return re.match("(X$|[A-Z][a-z]?-Bq|Bq)", self.element)

    def reset(self):
        if self.is_dummy:
            self._vdw = 0
            self._connectivity = 1000
            self._saturation = 0
            self._radii = 0
        else:
            self._set_radii()
            self._set_vdw()
            self._set_connectivity()
            self._set_saturation()

    def add_tag(self, *args):
        for a in args:
            if hasattr(a, "__iter__") and not isinstance(a, str):
                self.tags = self.tags.union(set(a))
            else:
                self.tags.add(a)
        return

    def get_invariant(self):
        """
        gets initial invariant
        (1) number of non-hydrogen connections (\d{1}): nconn
        (2) sum of bond order of non-hydrogen bonds * 10 (\d{2}): nB
        (3) atomic number (\d{3}): z
        #(4) sign of charge (\d{1})
        #(5) absolute charge (\d{1})
        (6) number of attached hydrogens (\d{1}): nH
        """
        heavy = set([x for x in self.connected if x.element != "H"])
        # number of non-hydrogen connections:
        nconn = len(heavy)
        # number of bonds with heavy atoms
        nB = 0
        for h in heavy:
            nB += BondOrder.get(h, self)
        # number of connected hydrogens
        nH = len(self.connected - heavy)
        # atomic number
        z = ELEMENTS.index(self.element)

        return "{:01d}{:03d}{:03d}{:01d}".format(
            int(nconn), int(nB * 10), int(z), int(nH)
        )

    def get_neighbor_id(self):
        """
        gets initial invariant based on self's element and the element of
        the atoms connected to self
        """
        # atomic number
        z = ELEMENTS.index(self.element)
        heavy = [
            ELEMENTS.index(x.element)
            for x in self.connected
            if x.element != "H"
        ]
        # number of non-hydrogen connections
        # number of bonds with heavy atoms and their element
        t = []
        for h in sorted(set(heavy)):
            t.extend([h, heavy.count(h)])

        # number of connected hydrogens
        nH = len(self.connected) - len(heavy)

        fmt = "%03i%02i" + (len(set(heavy)) * "%03i%02i") + "%02i"
        s = fmt % (z, len(heavy), *t, nH)

        return s

    def copy(self):
        rv = Atom()
        for key, val in self.__dict__.items():
            if key == "connected":
                continue
            if key == "constraint":
                continue
            if key == "_hashed":
                continue
            try:
                rv.__dict__[key] = val.copy()
            except AttributeError:
                rv.__dict__[key] = val
                # ignore chimerax objects so seqcrow doesn't print a
                # warning when a geometry is copied
                if "chimerax" in val.__class__.__module__:
                    continue
                if val.__class__.__module__ != "builtins":
                    self.LOG.warning(
                        "No copy method for {}: in-place changes may occur".format(
                            type(val)
                        )
                    )
        return rv

    # measurement
    def is_connected(self, other, tolerance=None):
        """determines if distance between atoms is small enough to be bonded"""
        return self.dist_is_connected(other, self.dist(other), tolerance)

    def dist_is_connected(self, other, dist_to_other, tolerance):
        """
        determines if distance between atoms is small enough to be bonded
        used to optimize connected checks when distances can be quickly precalculated
        like with scipy.spatial.distance_matrix
        """
        if tolerance is None:
            tolerance = 0.3

        if self._radii is None:
            self._set_radii()
        if other._radii is None:
            other._set_radii()
        cutoff = self._radii + other._radii + tolerance
        return dist_to_other < cutoff

    def add_bond_to(self, other):
        """add self and other to eachother's connected attribute"""
        self.connected.add(other)
        other.connected.add(self)

    def bond(self, other):
        """returns the vector self-->other"""
        return np.array(other.coords) - np.array(self.coords)

    def dist(self, other):
        """returns the distance between self and other"""
        return np.linalg.norm(self.bond(other))

    def angle(self, a1, a3):
        """returns the a1-self-a3 angle"""
        v1 = self.bond(a1)
        v2 = self.bond(a3)
        dot = np.dot(v1, v2)
        # numpy is still unhappy with this sometimes
        # every now and again, the changeElement cls test will "fail" b/c
        # numpy throws a warning here
        if abs(dot / (self.dist(a1) * self.dist(a3))) >= 1:
            return 0
        else:
            return np.arccos(dot / (self.dist(a1) * self.dist(a3)))

    def mass(self):
        """returns atomic mass"""
        if self.element in MASS:
            return MASS[self.element]
        elif not self.is_dummy:
            self.LOG.warning("no mass for %s" % self.element)
        return 0

    def rij(self, other):
        try:
            rv = RIJ[self.element + other.element]
        except KeyError:
            try:
                rv = RIJ[other.element + self.element]
            except KeyError:
                warn_LJ.add("".join(sorted([self.element, other.element])))
                return 0
        return rv

    def eij(self, other):
        try:
            rv = EIJ[self.element + other.element]
        except KeyError:
            try:
                rv = EIJ[other.element + self.element]
            except KeyError:
                warn_LJ.add("".join(sorted([self.element, other.element])))
                return 0
        return rv

    def bond_order(self, other):
        return BondOrder.get(self, other)

    @classmethod
    def get_shape(cls, shape_name):
        """returns dummy atoms in an idealized vsepr geometry
        shape_name can be:
        point
        linear 1
        linear 2
        bent 2 tetrahedral
        bent 2 planar
        trigonal planar
        bent 3 tetrahedral
        t shaped
        tetrahedral
        sawhorse
        seesaw
        square planar
        trigonal pyramidal
        trigonal bipyramidal
        square pyramidal
        pentagonal
        hexagonal
        trigonal prismatic
        pentagonal pyramidal
        octahedral
        capped octahedral
        hexagonal pyramidal
        pentagonal bipyramidal
        capped trigonal prismatic
        heptagonal
        hexagonal bipyramidal
        heptagonal pyramidal
        octagonal
        square antiprismatic
        trigonal dodecahedral
        capped cube
        biaugmented trigonal prismatic
        cubic
        elongated trigonal bipyramidal
        capped square antiprismatic
        enneagonal
        heptagonal bipyramidal
        hula-hoop
        triangular cupola
        tridiminished icosahedral
        muffin
        octagonal pyramidal
        tricapped trigonal prismatic
        """
        if shape_name == "point":
            return cls.linear_shape()[0:1]
        elif shape_name == "linear 1":
            return cls.linear_shape()[0:2]
        elif shape_name == "linear 2":
            return cls.linear_shape()
        elif shape_name == "bent 2 tetrahedral":
            return cls.tetrahedral_shape()[0:3]
        elif shape_name == "bent 2 planar":
            return cls.trigonal_planar_shape()[0:3]
        elif shape_name == "trigonal planar":
            return cls.trigonal_planar_shape()
        elif shape_name == "bent 3 tetrahedral":
            return cls.tetrahedral_shape()[0:4]
        elif shape_name == "t shaped":
            return cls.octahedral_shape()[0:4]
        elif shape_name == "tetrahedral":
            return cls.tetrahedral_shape()
        elif shape_name == "sawhorse":
            return (
                cls.trigonal_bipyramidal_shape()[0:3]
                + cls.trigonal_bipyramidal_shape()[-2:]
            )
        elif shape_name == "seesaw":
            return cls.octahedral_shape()[0:3] + cls.octahedral_shape()[-2:]
        elif shape_name == "square planar":
            return cls.octahedral_shape()[0:5]
        elif shape_name == "trigonal pyramidal":
            return cls.trigonal_bipyramidal_shape()[0:5]
        elif shape_name == "trigonal bipyramidal":
            return cls.trigonal_bipyramidal_shape()
        elif shape_name == "square pyramidal":
            return cls.octahedral_shape()[0:6]
        elif shape_name == "pentagonal":
            return cls.pentagonal_bipyraminal_shape()[0:6]
        elif shape_name == "hexagonal":
            return cls.hexagonal_bipyramidal_shape()[0:7]
        elif shape_name == "trigonal prismatic":
            return cls.trigonal_prismatic_shape()
        elif shape_name == "pentagonal pyramidal":
            return cls.pentagonal_bipyraminal_shape()[0:7]
        elif shape_name == "octahedral":
            return cls.octahedral_shape()
        elif shape_name == "capped octahedral":
            return cls.capped_octahedral_shape()
        elif shape_name == "hexagonal pyramidal":
            return cls.hexagonal_bipyramidal_shape()[0:8]
        elif shape_name == "pentagonal bipyramidal":
            return cls.pentagonal_bipyraminal_shape()
        elif shape_name == "capped trigonal prismatic":
            return cls.capped_trigonal_prismatic_shape()
        elif shape_name == "heptagonal":
            return cls.heptagonal_bipyramidal_shape()[0:8]
        elif shape_name == "hexagonal bipyramidal":
            return cls.hexagonal_bipyramidal_shape()
        elif shape_name == "heptagonal pyramidal":
            return cls.heptagonal_bipyramidal_shape()[0:9]
        elif shape_name == "octagonal":
            return cls.octagonal_pyramidal_shape()[0:9]
        elif shape_name == "square antiprismatic":
            return cls.square_antiprismatic_shape()
        elif shape_name == "trigonal dodecahedral":
            return cls.trigonal_dodecahedral_shape()
        elif shape_name == "capped cube":
            return cls.capped_cube_shape()
        elif shape_name == "biaugmented trigonal prismatic":
            return cls.biaugmented_trigonal_prismatic_shape()
        elif shape_name == "cubic":
            return cls.cubic_shape()
        elif shape_name == "elongated trigonal bipyramidal":
            return cls.elongated_trigonal_bipyramidal_shape()
        elif shape_name == "capped square antiprismatic":
            return cls.capped_square_antiprismatic_shape()
        elif shape_name == "enneagonal":
            return cls.enneagonal_shape()
        elif shape_name == "heptagonal bipyramidal":
            return cls.heptagonal_bipyramidal_shape()
        elif shape_name == "hula-hoop":
            return cls.hula_hoop_shape()
        elif shape_name == "triangular cupola":
            return cls.triangular_cupola_shape()
        elif shape_name == "tridiminished icosahedral":
            return cls.tridiminished_icosahedral_shape()
        elif shape_name == "muffin":
            return cls.muffin_shape()
        elif shape_name == "octagonal pyramidal":
            return cls.octagonal_pyramidal_shape()
        elif shape_name == "tricapped trigonal prismatic":
            return cls.tricapped_trigonal_prismatic_shape()
        else:
            raise RuntimeError(
                "no shape method is defined for %s" % shape_name
            )

    @classmethod
    def linear_shape(cls):
        """returns a list of 3 dummy atoms in a linear shape"""
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([1.0, 0.0, 0.0]), name="1")
        pos2 = Atom("X", np.array([-1.0, 0.0, 0.0]), name="2")

        return [center, pos1, pos2]

    @classmethod
    def trigonal_planar_shape(cls):
        """returns a list of 4 dummy atoms in a trigonal planar shape"""
        positions = cls.trigonal_bipyramidal_shape()
        return positions[:-2]

    @classmethod
    def tetrahedral_shape(cls):
        """returns a list of 5 dummy atoms in a tetrahedral shape"""
        center = Atom("X", np.zeros(3), name="0")
        angle = np.deg2rad(109.471 / 2)
        pos1 = Atom(
            "X", np.array([np.cos(angle), -np.sin(angle), 0.0]), name="1"
        )
        pos2 = Atom(
            "X", np.array([np.cos(angle), np.sin(angle), 0.0]), name="2"
        )
        pos3 = Atom(
            "X", np.array([-np.cos(angle), 0.0, np.sin(angle)]), name="3"
        )
        pos4 = Atom(
            "X", np.array([-np.cos(angle), 0.0, -np.sin(angle)]), name="4"
        )

        return [center, pos1, pos2, pos3, pos4]

    @classmethod
    def trigonal_bipyramidal_shape(cls):
        """returns a list of 6 dummy atoms in a trigonal bipyramidal shape"""
        center = Atom("X", np.zeros(3), name="0")
        angle = np.deg2rad(120)
        pos1 = Atom("X", np.array([1.0, 0.0, 0.0]), name="1")
        pos2 = Atom(
            "X", np.array([np.cos(angle), np.sin(angle), 0.0]), name="2"
        )
        pos3 = Atom(
            "X", np.array([np.cos(angle), -np.sin(angle), 0.0]), name="3"
        )
        pos4 = Atom("X", np.array([0.0, 0.0, 1.0]), name="4")
        pos5 = Atom("X", np.array([0.0, 0.0, -1.0]), name="5")

        return [center, pos1, pos2, pos3, pos4, pos5]

    @classmethod
    def octahedral_shape(cls):
        """returns a list of 7 dummy atoms in an octahedral shape"""
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([1.0, 0.0, 0.0]), name="1")
        pos2 = Atom("X", np.array([0.0, 1.0, 0.0]), name="2")
        pos3 = Atom("X", np.array([-1.0, 0.0, 0.0]), name="3")
        pos4 = Atom("X", np.array([0.0, -1.0, 0.0]), name="4")
        pos5 = Atom("X", np.array([0.0, 0.0, 1.0]), name="5")
        pos6 = Atom("X", np.array([0.0, 0.0, -1.0]), name="6")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6]

    @classmethod
    def trigonal_prismatic_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([-0.6547, -0.3780, 0.6547]), name="1")
        pos2 = Atom("X", np.array([-0.6547, -0.3780, -0.6547]), name="2")
        pos3 = Atom("X", np.array([0.6547, -0.3780, 0.6547]), name="3")
        pos4 = Atom("X", np.array([0.6547, -0.3780, -0.6547]), name="4")
        pos5 = Atom("X", np.array([0.0, 0.7559, 0.6547]), name="5")
        pos6 = Atom("X", np.array([0.0, 0.7559, -0.6547]), name="6")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6]

    @classmethod
    def capped_octahedral_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([0.0, 0.0, 1.0]), name="1")
        pos2 = Atom("X", np.array([0.9777, 0.0, 0.2101]), name="2")
        pos3 = Atom("X", np.array([0.1698, 0.9628, 0.2101]), name="3")
        pos4 = Atom("X", np.array([-0.9187, 0.3344, 0.2102]), name="4")
        pos5 = Atom("X", np.array([-0.4888, -0.8467, 0.2102]), name="5")
        pos6 = Atom("X", np.array([0.3628, -0.6284, -0.6881]), name="6")
        pos7 = Atom("X", np.array([-0.2601, 0.4505, -0.8540]), name="7")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7]

    @classmethod
    def capped_trigonal_prismatic_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([0.0, 0.0, 1.0]), name="1")
        pos2 = Atom("X", np.array([0.6869, 0.6869, 0.2374]), name="2")
        pos3 = Atom("X", np.array([-0.6869, 0.6869, 0.2374]), name="3")
        pos4 = Atom("X", np.array([0.6869, -0.6869, 0.2374]), name="4")
        pos5 = Atom("X", np.array([-0.6869, -0.6869, 0.2374]), name="5")
        pos6 = Atom("X", np.array([0.6175, 0.0, -0.7866]), name="6")
        pos7 = Atom("X", np.array([-0.6175, 0.0, -0.7866]), name="7")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7]

    @classmethod
    def pentagonal_bipyraminal_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([1.0, 0.0, 0.0]), name="1")
        pos2 = Atom("X", np.array([0.3090, 0.9511, 0.0]), name="2")
        pos3 = Atom("X", np.array([-0.8090, 0.5878, 0.0]), name="3")
        pos4 = Atom("X", np.array([-0.8090, -0.5878, 0.0]), name="4")
        pos5 = Atom("X", np.array([0.3090, -0.9511, 0.0]), name="5")
        pos6 = Atom("X", np.array([0.0, 0.0, 1.0]), name="6")
        pos7 = Atom("X", np.array([0.0, 0.0, -1.0]), name="7")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7]

    @classmethod
    def biaugmented_trigonal_prismatic_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([-0.6547, -0.3780, 0.6547]), name="1")
        pos2 = Atom("X", np.array([-0.6547, -0.3780, -0.6547]), name="2")
        pos3 = Atom("X", np.array([0.6547, -0.3780, 0.6547]), name="3")
        pos4 = Atom("X", np.array([0.6547, -0.3780, -0.6547]), name="4")
        pos5 = Atom("X", np.array([0.0, 0.7559, 0.6547]), name="5")
        pos6 = Atom("X", np.array([0.0, 0.7559, -0.6547]), name="6")
        pos7 = Atom("X", np.array([0.0, -1.0, 0.0]), name="7")
        pos8 = Atom("X", np.array([-0.8660, 0.5, 0.0]), name="8")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7, pos8]

    @classmethod
    def cubic_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([0.5775, 0.5774, 0.5774]), name="1")
        pos2 = Atom("X", np.array([0.5775, 0.5774, -0.5774]), name="2")
        pos3 = Atom("X", np.array([0.5775, -0.5774, 0.5774]), name="3")
        pos4 = Atom("X", np.array([-0.5775, 0.5774, 0.5774]), name="4")
        pos5 = Atom("X", np.array([0.5775, -0.5774, -0.5774]), name="5")
        pos6 = Atom("X", np.array([-0.5775, 0.5774, -0.5774]), name="6")
        pos7 = Atom("X", np.array([-0.5775, -0.5774, 0.5774]), name="7")
        pos8 = Atom("X", np.array([-0.5775, -0.5774, -0.5774]), name="8")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7, pos8]

    @classmethod
    def elongated_trigonal_bipyramidal_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([0.6547, 0.0, 0.7559]), name="1")
        pos2 = Atom("X", np.array([-0.6547, 0.0, 0.7559]), name="2")
        pos3 = Atom("X", np.array([0.6547, 0.6547, -0.3780]), name="3")
        pos4 = Atom("X", np.array([-0.6547, 0.6547, -0.3780]), name="4")
        pos5 = Atom("X", np.array([0.6547, -0.6547, -0.3780]), name="5")
        pos6 = Atom("X", np.array([-0.6547, -0.6547, -0.3780]), name="6")
        pos7 = Atom("X", np.array([1.0, 0.0, 0.0]), name="7")
        pos8 = Atom("X", np.array([-1.0, 0.0, 0.0]), name="8")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7, pos8]

    @classmethod
    def hexagonal_bipyramidal_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([0.0, -1.0, 0.0]), name="1")
        pos2 = Atom("X", np.array([0.8660, -0.5, 0.0]), name="2")
        pos3 = Atom("X", np.array([0.8660, 0.5, 0.0]), name="3")
        pos4 = Atom("X", np.array([0.0, 1.0, 0.0]), name="4")
        pos5 = Atom("X", np.array([-0.8660, 0.5, 0.0]), name="5")
        pos6 = Atom("X", np.array([-0.8660, -0.5, 0.0]), name="6")
        pos7 = Atom("X", np.array([0.0, 0.0, 1.0]), name="7")
        pos8 = Atom("X", np.array([0.0, 0.0, -1.0]), name="8")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7, pos8]

    @classmethod
    def square_antiprismatic_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([0.0, 0.0, 1.0]), name="1")
        pos2 = Atom("X", np.array([0.9653, 0.0, 0.2612]), name="2")
        pos3 = Atom("X", np.array([-0.5655, 0.7823, 0.2612]), name="3")
        pos4 = Atom("X", np.array([-0.8825, -0.3912, 0.2612]), name="4")
        pos5 = Atom("X", np.array([0.1999, -0.9444, 0.2612]), name="5")
        pos6 = Atom("X", np.array([0.3998, 0.7827, -0.4776]), name="6")
        pos7 = Atom("X", np.array([-0.5998, 0.1620, -0.7836]), name="7")
        pos8 = Atom("X", np.array([0.4826, -0.3912, -0.7836]), name="8")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7, pos8]

    @classmethod
    def trigonal_dodecahedral_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([-0.5997, 0.0, 0.8002]), name="1")
        pos2 = Atom("X", np.array([0.0, -0.9364, 0.3509]), name="2")
        pos3 = Atom("X", np.array([0.5998, 0.0, 0.8002]), name="3")
        pos4 = Atom("X", np.array([0.0, 0.9364, 0.3509]), name="4")
        pos5 = Atom("X", np.array([-0.9364, 0.0, -0.3509]), name="5")
        pos6 = Atom("X", np.array([0.0, -0.5997, -0.8002]), name="6")
        pos7 = Atom("X", np.array([0.9365, 0.0, -0.3509]), name="7")
        pos8 = Atom("X", np.array([0.0, 0.5997, -0.8002]), name="8")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7, pos8]

    @classmethod
    def heptagonal_bipyramidal_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([1.0, 0.0, 0.0]), name="1")
        pos2 = Atom("X", np.array([0.6235, 0.7818, 0.0]), name="2")
        pos3 = Atom("X", np.array([-0.2225, 0.9749, 0.0]), name="3")
        pos4 = Atom("X", np.array([-0.9010, 0.4339, 0.0]), name="4")
        pos5 = Atom("X", np.array([-0.9010, -0.4339, 0.0]), name="5")
        pos6 = Atom("X", np.array([-0.2225, -0.9749, 0.0]), name="6")
        pos7 = Atom("X", np.array([0.6235, -0.7818, 0.0]), name="7")
        pos8 = Atom("X", np.array([0.0, 0.0, 1.0]), name="8")
        pos9 = Atom("X", np.array([0.0, 0.0, -1.0]), name="9")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7, pos8, pos9]

    @classmethod
    def capped_cube_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([0.6418, 0.6418, 0.4196]), name="1")
        pos2 = Atom("X", np.array([0.6418, -0.6418, 0.4196]), name="2")
        pos3 = Atom("X", np.array([-0.6418, 0.6418, 0.4196]), name="3")
        pos4 = Atom("X", np.array([-0.6418, -0.6418, 0.4196]), name="4")
        pos5 = Atom("X", np.array([0.5387, 0.5387, -0.6478]), name="5")
        pos6 = Atom("X", np.array([0.5387, -0.5387, -0.6478]), name="6")
        pos7 = Atom("X", np.array([-0.5387, 0.5387, -0.6478]), name="7")
        pos8 = Atom("X", np.array([-0.5387, -0.5387, -0.6478]), name="8")
        pos9 = Atom("X", np.array([0.0, 0.0, 1.0]), name="9")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7, pos8, pos9]

    @classmethod
    def capped_square_antiprismatic_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([0.9322, 0.0, 0.3619]), name="1")
        pos2 = Atom("X", np.array([-0.9322, 0.0, 0.3619]), name="2")
        pos3 = Atom("X", np.array([0.0, 0.9322, 0.3619]), name="3")
        pos4 = Atom("X", np.array([0.0, -0.9322, 0.3619]), name="4")
        pos5 = Atom("X", np.array([0.5606, 0.5606, -0.6095]), name="5")
        pos6 = Atom("X", np.array([-0.5606, 0.5606, -0.6095]), name="6")
        pos7 = Atom("X", np.array([-0.5606, -0.5606, -0.6095]), name="7")
        pos8 = Atom("X", np.array([0.5606, -0.5606, -0.6095]), name="8")
        pos9 = Atom("X", np.array([0.0, 0.0, 1.0]), name="9")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7, pos8, pos9]

    @classmethod
    def enneagonal_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([1.0, 0.0, 0.0]), name="1")
        pos2 = Atom("X", np.array([0.7660, 0.6428, 0.0]), name="2")
        pos3 = Atom("X", np.array([0.1736, 0.9848, 0.0]), name="3")
        pos4 = Atom("X", np.array([-0.5, 0.8660, 0.0]), name="4")
        pos5 = Atom("X", np.array([-0.9397, 0.3420, 0.0]), name="5")
        pos6 = Atom("X", np.array([-0.9397, -0.3420, 0.0]), name="6")
        pos7 = Atom("X", np.array([-0.5, -0.8660, 0.0]), name="7")
        pos8 = Atom("X", np.array([0.1736, -0.9848, 0.0]), name="8")
        pos9 = Atom("X", np.array([0.7660, -0.6428, 0.0]), name="9")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7, pos8, pos9]

    @classmethod
    def hula_hoop_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([1.0, 0.0, 0.0]), name="1")
        pos2 = Atom("X", np.array([0.5, 0.8660, 0.0]), name="2")
        pos3 = Atom("X", np.array([-0.5, 0.8660, 0.0]), name="3")
        pos4 = Atom("X", np.array([-1.0, 0.0, 0.0]), name="4")
        pos5 = Atom("X", np.array([-0.5, -0.8660, 0.0]), name="5")
        pos6 = Atom("X", np.array([0.5, -0.8660, 0.0]), name="6")
        pos7 = Atom("X", np.array([0.0, 0.0, 1.0]), name="7")
        pos8 = Atom("X", np.array([0.5, 0.0, -0.8660]), name="8")
        pos9 = Atom("X", np.array([-0.5, 0.0, -0.8660]), name="9")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7, pos8, pos9]

    @classmethod
    def triangular_cupola_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([1.0, 0.0, 0.0]), name="1")
        pos2 = Atom("X", np.array([0.5, 0.8660, 0.0]), name="2")
        pos3 = Atom("X", np.array([-0.5, 0.8660, 0.0]), name="3")
        pos4 = Atom("X", np.array([-1.0, 0.0, 0.0]), name="4")
        pos5 = Atom("X", np.array([0.5, -0.8660, 0.0]), name="5")
        pos6 = Atom("X", np.array([-0.5, -0.8660, 0.0]), name="6")
        pos7 = Atom("X", np.array([0.5, 0.2887, -0.8165]), name="7")
        pos8 = Atom("X", np.array([-0.5, 0.2887, -0.8165]), name="8")
        pos9 = Atom("X", np.array([0.0, -0.5774, -0.8165]), name="9")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7, pos8, pos9]

    @classmethod
    def tridiminished_icosahedral_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([-0.2764, 0.8507, -0.4472]), name="1")
        pos2 = Atom("X", np.array([-0.8944, 0.0, -0.4472]), name="2")
        pos3 = Atom("X", np.array([-0.2764, -0.8507, -0.4472]), name="3")
        pos4 = Atom("X", np.array([0.7236, -0.5257, -0.4472]), name="4")
        pos5 = Atom("X", np.array([0.8944, 0.0, 0.4472]), name="5")
        pos6 = Atom("X", np.array([0.2764, 0.8507, 0.4472]), name="6")
        pos7 = Atom("X", np.array([-0.7236, -0.5257, 0.4472]), name="7")
        pos8 = Atom("X", np.array([0.0, 0.0, 1.0]), name="8")
        pos9 = Atom("X", np.array([0.0, 0.0, -1.0]), name="9")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7, pos8, pos9]

    @classmethod
    def muffin_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([0.0, 0.9875, 0.1579]), name="1")
        pos2 = Atom("X", np.array([0.9391, 0.3051, 0.1579]), name="2")
        pos3 = Atom("X", np.array([0.5804, -0.7988, 0.1579]), name="3")
        pos4 = Atom("X", np.array([-0.5804, -0.7988, 0.1579]), name="4")
        pos5 = Atom("X", np.array([-0.9391, 0.3055, 0.1579]), name="5")
        pos6 = Atom("X", np.array([-0.5799, -0.3356, -0.7423]), name="6")
        pos7 = Atom("X", np.array([0.5799, -0.3356, -0.7423]), name="7")
        pos8 = Atom("X", np.array([0.0, 0.6694, -0.7429]), name="8")
        pos9 = Atom("X", np.array([0.0, 0.0, 1.0]), name="9")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7, pos8, pos9]

    @classmethod
    def octagonal_pyramidal_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([0.7071, -0.7071, 0.0]), name="1")
        pos2 = Atom("X", np.array([1.0, 0.0, 0.0]), name="2")
        pos3 = Atom("X", np.array([0.7071, 0.7071, 0.0]), name="3")
        pos4 = Atom("X", np.array([0.0, 1.0, 0.0]), name="4")
        pos5 = Atom("X", np.array([-0.7071, 0.7071, 0.0]), name="5")
        pos6 = Atom("X", np.array([-1.0, 0.0, 0.0]), name="6")
        pos7 = Atom("X", np.array([-0.7071, -0.7071, 0.0]), name="7")
        pos8 = Atom("X", np.array([0.0, -1.0, 0.0]), name="8")
        pos9 = Atom("X", np.array([0.0, 0.0, -1.0]), name="9")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7, pos8, pos9]

    @classmethod
    def tricapped_trigonal_prismatic_shape(cls):
        center = Atom("X", np.zeros(3), name="0")
        pos1 = Atom("X", np.array([0.0, 0.0, 1.0]), name="1")
        pos2 = Atom("X", np.array([-0.2357, 0.9129, 0.3333]), name="2")
        pos3 = Atom("X", np.array([-0.9428, 0.0, 0.3333]), name="3")
        pos4 = Atom("X", np.array([0.2357, -0.9129, 0.3333]), name="4")
        pos5 = Atom("X", np.array([0.9428, 0.0, 0.3333]), name="5")
        pos6 = Atom("X", np.array([0.5303, 0.6847, -0.5]), name="6")
        pos7 = Atom("X", np.array([-0.5303, -0.6847, -0.5]), name="7")
        pos8 = Atom("X", np.array([-0.5893, 0.4564, -0.6667]), name="8")
        pos9 = Atom("X", np.array([0.5893, -0.4564, -0.6667]), name="9")

        return [center, pos1, pos2, pos3, pos4, pos5, pos6, pos7, pos8, pos9]

    @staticmethod
    def new_shape(old_shape, new_connectivity, bond_change):
        """returns the name of the expected vsepr geometry when the number of bonds
        changes by +/- 1

        old_shape - :str: vsepr geometry name
        new_connectivity - :int: connectivity (see Atom._connectivity)
        bond_change - :int: +1 or -1, indicating that the number of bonds is changing by 1"""
        if old_shape == "point":
            if bond_change == 1:
                return "linear 1"
            else:
                return None

        elif old_shape == "linear 1":
            if bond_change == 1:
                return "linear 2"
            elif bond_change == -1:
                return None

        elif old_shape == "linear 2":
            if bond_change == 1:
                if new_connectivity is not None and new_connectivity > 4:
                    return "t shaped"
                else:
                    return "trigonal planar"
            elif bond_change == -1:
                return "linear 1"

        elif old_shape == "bent 2 tetrahedral":
            if bond_change == 1:
                return "bent 3 tetrahedral"
            elif bond_change == -1:
                return "linear 1"

        elif old_shape == "bent 2 planar":
            if bond_change == 1:
                return "trigonal planar"
            elif bond_change == -1:
                return "linear 1"

        elif old_shape == "trigonal planar":
            if bond_change == 1:
                return "tetrahedral"
            elif bond_change == -1:
                if new_connectivity == 4:
                    return "bent 2 tetrahedral"
                else:
                    return "bent 2 planar"

        elif old_shape == "bent 3 tetrahedral":
            if bond_change == 1:
                return "tetrahedral"
            elif bond_change == -1:
                return "bent 2 tetrahedral"

        elif old_shape == "t shaped":
            if bond_change == 1:
                if new_connectivity == 6:
                    return "square planar"
                else:
                    return "sawhorse"
            elif bond_change == -1:
                return "linear 2"

        elif old_shape == "tetrahedral":
            if bond_change == 1:
                return "trigonal bipyramidal"
            elif bond_change == -1:
                return "bent 3 tetrahedral"

        elif old_shape == "square planar":
            if bond_change == 1:
                return "trigonal bipyramidal"
            elif bond_change == -1:
                return "t shaped"

        elif old_shape == "trigonal bipyramidal":
            if bond_change == 1:
                return "octahedral"
            elif bond_change == -1:
                return "sawhorse"

        elif old_shape == "octahedral":
            if bond_change == -1:
                return "trigonal bipyramid"

        else:
            raise RuntimeError("no shape method is defined for %s" % old_shape)

    def get_vsepr(self):
        """determine vsepr geometry around an atom
        returns shape as a string and the score assigned to that shape
        returns None if self has > 6 bonds
        scores > 0.5 are generally questionable
        see atom.get_shape for a list of shapes
        """

        # shapes with a code in the commend next to them are from Simas et. al. Inorg. Chem. 2018, 57, 17, 10557–10567

        # determine what geometries to try based on the number of bonded atoms
        try_shapes = {}
        if len(self.connected) == 0:
            try_shapes["point"] = Atom.get_shape("point")

        elif len(self.connected) == 1:
            try_shapes["linear 1"] = Atom.get_shape("linear 1")

        elif len(self.connected) == 2:
            try_shapes["linear 2"] = Atom.get_shape("linear 2")
            try_shapes["bent 2 planar"] = Atom.get_shape("bent 2 planar")
            try_shapes["bent 2 tetrahedral"] = Atom.get_shape(
                "bent 2 tetrahedral"
            )

        elif len(self.connected) == 3:
            try_shapes["trigonal planar"] = Atom.get_shape("trigonal planar")
            try_shapes["bent 3 tetrahedral"] = Atom.get_shape(
                "bent 3 tetrahedral"
            )
            try_shapes["t shaped"] = Atom.get_shape("t shaped")

        elif len(self.connected) == 4:
            try_shapes["tetrahedral"] = Atom.get_shape("tetrahedral")
            try_shapes["sawhorse"] = Atom.get_shape("sawhorse")
            try_shapes["seesaw"] = Atom.get_shape("seesaw")
            try_shapes["square planar"] = Atom.get_shape("square planar")
            try_shapes["trigonal pyramidal"] = Atom.get_shape(
                "trigonal pyramidal"
            )

        elif len(self.connected) == 5:
            try_shapes["trigonal bipyramidal"] = Atom.get_shape(
                "trigonal bipyramidal"
            )
            try_shapes["square pyramidal"] = Atom.get_shape("square pyramidal")
            try_shapes["pentagonal"] = Atom.get_shape("pentagonal")  # PP-5

        elif len(self.connected) == 6:
            try_shapes["octahedral"] = Atom.get_shape("octahedral")
            try_shapes["hexagonal"] = Atom.get_shape("hexagonal")  # HP-6
            try_shapes["trigonal prismatic"] = Atom.get_shape(
                "trigonal prismatic"
            )  # TPR-6
            try_shapes["pentagonal pyramidal"] = Atom.get_shape(
                "pentagonal pyramidal"
            )  # PPY-6
        elif len(self.connected) == 7:
            try_shapes["capped octahedral"] = Atom.get_shape(
                "capped octahedral"
            )  # COC-7
            try_shapes["capped trigonal prismatic"] = Atom.get_shape(
                "capped trigonal prismatic"
            )  # CTPR-7
            try_shapes["heptagonal"] = Atom.get_shape("heptagonal")  # HP-7
            try_shapes["hexagonal pyramidal"] = Atom.get_shape(
                "hexagonal pyramidal"
            )  # HPY-7
            try_shapes["pentagonal bipyramidal"] = Atom.get_shape(
                "pentagonal bipyramidal"
            )  # PBPY-7
        elif len(self.connected) == 8:
            try_shapes["biaugmented trigonal prismatic"] = Atom.get_shape(
                "biaugmented trigonal prismatic"
            )  # BTPR-8
            try_shapes["cubic"] = Atom.get_shape("cubic")  # CU-8
            try_shapes["elongated trigonal bipyramidal"] = Atom.get_shape(
                "elongated trigonal bipyramidal"
            )  # ETBPY-8
            try_shapes["hexagonal bipyramidal"] = Atom.get_shape(
                "hexagonal bipyramidal"
            )  # HBPY-8
            try_shapes["heptagonal pyramidal"] = Atom.get_shape(
                "heptagonal pyramidal"
            )  # HPY-8
            try_shapes["octagonal"] = Atom.get_shape("octagonal")  # OP-8
            try_shapes["square antiprismatic"] = Atom.get_shape(
                "square antiprismatic"
            )  # SAPR-8
            try_shapes["trigonal dodecahedral"] = Atom.get_shape(
                "trigonal dodecahedral"
            )  # TDD-8
        elif len(self.connected) == 9:
            try_shapes["capped cube"] = Atom.get_shape("capped cube")  # CCU-9
            try_shapes["capped square antiprismatic"] = Atom.get_shape(
                "capped square antiprismatic"
            )  # CSAPR-9
            try_shapes["enneagonal"] = Atom.get_shape("enneagonal")  # EP-9
            try_shapes["heptagonal bipyramidal"] = Atom.get_shape(
                "heptagonal bipyramidal"
            )  # HBPY-9
            try_shapes["hula-hoop"] = Atom.get_shape("hula-hoop")  # HH-9
            try_shapes["triangular cupola"] = Atom.get_shape(
                "triangular cupola"
            )  # JTC-9
            try_shapes["tridiminished icosahedral"] = Atom.get_shape(
                "tridiminished icosahedral"
            )  # JTDIC-9
            try_shapes["muffin"] = Atom.get_shape("muffin")  # MFF-9
            try_shapes["octagonal pyramidal"] = Atom.get_shape(
                "octagonal pyramidal"
            )  # OPY-9
            try_shapes["tricapped trigonal prismatic"] = Atom.get_shape(
                "tricapped trigonal prismatic"
            )  # TCTPR-9

        else:
            return None, None

        # make a copy of the atom and the atoms bonded to it
        # set each bond length to 1 to more easily compare to the
        # idealized shapes from Atom
        adjusted_shape = [atom.copy() for atom in [self, *self.connected]]
        for atom in adjusted_shape:
            atom.coords -= self.coords

        for atom in adjusted_shape[1:]:
            atom.coords /= atom.dist(adjusted_shape[0])

        Y = np.array([position.coords for position in adjusted_shape])
        r1 = np.matmul(np.transpose(Y), Y)
        u, s1, vh = np.linalg.svd(r1)

        best_score = None
        best_shape = None
        for shape in try_shapes:
            X = np.array([position.coords for position in try_shapes[shape]])
            r2 = np.matmul(np.transpose(X), X)
            u, s2, vh = np.linalg.svd(r2)

            score = sum([abs(x1 - x2) for x1, x2 in zip(s1, s2)])
            if best_score is None or score < best_score:
                best_score = score
                best_shape = shape

        return best_shape, best_score



#!/usr/bin/env python3

import sys
import argparse

from AaronTools.geometry import Geometry
from AaronTools.fileIO import FileReader, read_types
from AaronTools.utils.utils import get_filename, glob_files

def two_atoms_and_a_float(vals):
    """check to see if argument is two ints and a float"""
    err = argparse.ArgumentTypeError(
        "error with atom/distance specification: %s\n" % " ".join(vals) +
        "expected -s/-c int int float\nexample: -c 5 6 +9001"
    )
    out = []
    if len(vals) != 3:
        raise err
    for v in vals[:-1]:
        try:
            out.append(int(v)-1)
        except ValueError:
            raise err

        if int(v) != float(v):
            raise err

    try:
        out.append(float(vals[-1]))
    except ValueError:
        raise err

    return out


bond_parser = argparse.ArgumentParser(
    description="measure or modify distance between atoms",
    formatter_class=argparse.RawTextHelpFormatter
)

bond_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file"
)

bond_parser.add_argument(
    "-if", "--input-format",
    type=str,
    default=None,
    choices=read_types,
    dest="input_format",
    help="file format of input - xyz is assumed if input is stdin"
)

bond_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    required=False,
    dest="outfile",
    metavar="output destination",
    help="output destination\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: stdout"
)

bond_parser.add_argument(
    "-m", "--measure",
    metavar=("atom1", "atom2"),
    action="append",
    type=int,
    nargs=2,
    default=[],
    required=False,
    dest="measure",
    help="measure and print distance between atoms (1-indexed)"
)

bond_parser.add_argument(
    "-c", "--change",
    metavar=("atom1", "atom2", "increment"),
    action="append",
    type=str,
    nargs=3,
    default=[],
    required=False,
    dest="change",
    help="change distance by the amount specified"
)

bond_parser.add_argument(
    "-s", "--set", metavar=("atom1", "atom2", "distance"),
    action="append",
    type=str,
    nargs=3,
    default=[],
    required=False,
    dest="set_ang",
    help="set distance to the amount specified"
)


args = bond_parser.parse_args()

for f in glob_files(args.infile, parser=bond_parser):
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format, None))
        else:
            infile = FileReader(f)
    else:
        if args.input_format is not None:
            infile = FileReader(("from stdin", args.input_format, f))
        else:
            infile = FileReader(("from stdin", "xyz", f))

    geom = Geometry(infile)

    #set bond to specified value
    for bond in args.set_ang:
        vals = two_atoms_and_a_float(bond)
        a1 = geom.find(str(vals[0]))[0]
        a2 = geom.find(str(vals[1]))[0]
        geom.change_distance(a1, a2, dist=vals[2], adjust=False)

    #change bond by specified amount
    for bond in args.change:
        vals = two_atoms_and_a_float(bond)
        a1 = geom.find(str(vals[0]))[0]
        a2 = geom.find(str(vals[1]))[0]
        geom.change_distance(a1, a2, dist=vals[2], adjust=True)

    #print specified bonds
    out = ""
    for bond in args.measure:
        a1 = geom.find(str(bond[0]))[0]
        a2 = geom.find(str(bond[1]))[0]
        val = a1.dist(a2)
        out += "%f\n" % val

    out = out.rstrip()

    if len(args.set_ang) + len(args.change) > 0:
        if args.outfile:
            outfile = args.outfile
            if "$INFILE" in outfile:
                outfile = outfile.replace("$INFILE", get_filename(f))
            geom.write(append=True, outfile=outfile)
        else:
            print(geom.write(outfile=False))

    if len(args.measure) > 0:
        if not args.outfile:
            print(out)
        else:
            outfile = args.outfile
            if "$INFILE" in outfile:
                outfile = outfile.replace("$INFILE", get_filename(f))
            with open(outfile, "a") as f:
                f.write(out)



#!/usr/bin/env python3

import argparse
from sys import stdin

from AaronTools.fileIO import FileReader, read_types
from AaronTools.geometry import Geometry
from AaronTools.ring import Ring
from AaronTools.utils.utils import get_filename, glob_files

ring_parser = argparse.ArgumentParser(
    description="close rings on a geometry",
    formatter_class=argparse.RawTextHelpFormatter
)

ring_parser.add_argument(
    "infile",
    metavar="input file",
    type=str,
    nargs="*",
    default=[stdin],
    help="a coordinate file"
)

ring_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    required=False,
    metavar="output destination",
    dest="outfile",
    help="output destination\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: stdout"
)

ring_parser.add_argument(
    "-ls", "--list",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="list_avail",
    help="list available rings"
)

ring_parser.add_argument(
    "-if", "--input-format",
    type=str,
    nargs=1,
    default=None,
    choices=read_types,
    dest="input_format",
    help="file format of input - xyz is assumed if input is stdin"
)

ring_parser.add_argument(
    "-r", "--ring",
    metavar=("atom1", "atom2", "ring"),
    type=str,
    nargs=3,
    action="append",
    default=None,
    required=False,
    dest="substitutions",
    help="substitution instructions \n" +
    "atom1 and atom2 specify the position to add the new ring"
)

ring_parser.add_argument(
    "-m", "--minimize",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="minimize",
    help="try to minimize structure difference"
)

ring_parser.add_argument(
    "-f", "--flip-rings",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="flip",
    help="also try swapping target order when minimizing"
)

args = ring_parser.parse_args()

if args.list_avail:
    s = ""
    for i, name in enumerate(sorted(Ring.list())):
        s += "%-20s" % name
        # if (i + 1) % 3 == 0:
        if (i + 1) % 1 == 0:
            s += "\n"

    print(s.strip())
    exit(0)

for infile in glob_files(args.infile, parser=ring_parser):
    if isinstance(infile, str):
        if args.input_format is not None:
            f = FileReader((infile, args.input_format, infile))
        else:
            f = FileReader(infile)
    else:
        if args.input_format is not None:
            f = FileReader(("from stdin", args.input_format, stdin))
        else:
            f = FileReader(("from stdin", "xyz", stdin))

    geom = Geometry(f)

    targets = {}

    for sub_info in args.substitutions:
        ring = sub_info[2]

        ring_geom = Ring(ring)

        key = ",".join(sub_info[:2])
        if key in targets:
            targets[key].append(ring_geom)
        else:
            targets[key] = [ring_geom]

    for key in targets:
        for ring_geom in targets[key]:
            geom.ring_substitute(
                key,
                ring_geom,
                minimize=args.minimize,
                flip_walk=args.flip,
            )

    if args.outfile:
        outfile = args.outfile
        if "$INFILE" in outfile:
            outfile = outfile.replace("$INFILE", get_filename(infile))
        geom.write(append=True, outfile=outfile)
    else:
        print(geom.write(outfile=False))



#! /usr/bin/env python3
"""Testing for getting geometries, substituents, and rings from
an IUPAC name or a SMILES string"""
import os
import unittest

import numpy as np
from AaronTools.geometry import Geometry
from AaronTools.ring import Ring
from AaronTools.substituent import Substituent
from AaronTools.test import TestWithTimer, prefix, rmsd_tol, validate


class TestFromString(TestWithTimer):
    """
    these only get run if RDKit is installed
    molecules can be fetched from CATCVS, but there can be
    discrepencies that make these tests "fail"
    """

    COCH3 = Substituent("COCH3")
    NO2 = Substituent("NO2")
    benzene = Ring("benzene")
    chiral_geom = Geometry(
        os.path.join(prefix, "test_files", "chiral_ring.xyz")
    )

    def is_COCH3(self, sub, thresh=0.03):
        ref = TestFromString.COCH3
        ref.refresh_connected()
        sub.refresh_connected()
        ref.refresh_ranks()
        ref.refresh_ranks()
        ref.atoms = ref.reorder(start=ref.atoms[0])[0]
        sub.atoms = sub.reorder(start=sub.atoms[0])[0]
        self.assertTrue(
            validate(
                sub,
                ref,
                thresh=thresh,
                heavy_only=True,
                sort=False,
                debug=False,
            )
        )

    def is_NO2(self, sub):
        ref = TestFromString.NO2
        self.assertTrue(validate(sub, ref, thresh=2e-1))

    def test_substituent(self):
        try:
            import rdkit

            sub = Substituent.from_string(
                "acetyl", form="iupac", strict_use_rdkit=True
            )
            self.is_COCH3(sub)

            with self.assertLogs(Substituent.LOG, level="WARNING"):
                sub = Substituent.from_string(
                    "nitro", form="iupac", strict_use_rdkit=True
                )
                self.is_NO2(sub)

            sub = Substituent.from_string(
                "O=[N.]=O", form="smiles", strict_use_rdkit=True
            )
            self.is_NO2(sub)

            sub = Substituent.from_string(
                "O=[N]=O", form="smiles", strict_use_rdkit=True
            )
            self.is_NO2(sub)

        except (ImportError, ModuleNotFoundError):
            # I still want to test CACTVS things because sometimes they change stuff
            # that breaks our stuff
            if any(
                user == os.getenv("USER", os.getenv("USERNAME", False))
                for user in ["ajs99778", "normn"]
            ):
                sub = Substituent.from_string("acetyl", form="iupac")
                print(sub.write(outfile=False))
                self.is_COCH3(sub, thresh=0.3)

                sub = Substituent.from_string("nitro", form="iupac")
                print(sub.write(outfile=False))
                self.is_NO2(sub)

                sub = Substituent.from_string("O=[N.]=O", form="smiles")
                print(sub.write(outfile=False))
                self.is_NO2(sub)

                sub = Substituent.from_string("O=[N]=O", form="smiles")
                print(sub.write(outfile=False))
                self.is_NO2(sub)

            else:
                self.skipTest("RDKit not installed, CACTVS is not tested")

    def test_geometry(self):
        try:
            import rdkit

            geom = Geometry.from_string(
                "(1R,2R)-1-Chloro-2-methylcyclohexane", form="iupac"
            )
            ref = TestFromString.chiral_geom
            # really loose threshhold b/c rdkit can give a boat cyclohexane...
            self.assertTrue(validate(geom, ref, thresh=0.35, heavy_only=True))

        except (ImportError, ModuleNotFoundError):
            if any(
                user == os.getenv("USER", os.getenv("USERNAME", False))
                for user in ["ajs99778", "normn"]
            ):
                geom = Geometry.from_string(
                    "(1R,2R)-1-Chloro-2-methylcyclohexane", form="iupac"
                )
                ref = TestFromString.chiral_geom
                # really loose threshhold b/c rdkit can give a boat cyclohexane...
                self.assertTrue(
                    validate(geom, ref, thresh=0.35, heavy_only=True)
                )
            else:
                self.skipTest("RDKit not installed, CACTVS is not tested")

    def test_ring(self):
        try:
            import rdkit

            ring = Ring.from_string(
                "benzene", end_length=1, end_atom="C", form="iupac"
            )
            ref = self.benzene
            self.assertTrue(validate(ring, ref, thresh="loose"))

        except ImportError:
            if any(
                user == os.getenv("USER", os.getenv("USERNAME", False))
                for user in ["ajs99778", "normn"]
            ):
                ring = Ring.from_string(
                    "benzene", end_length=1, end_atom="C", form="iupac"
                )
                print(ring.comment)
                ref = self.benzene
                self.assertTrue(
                    validate(ring, ref, thresh="loose", debug=True)
                )

            else:
                self.skipTest("RDKit not installed, CACTVS is not tested")


if __name__ == "__main__":
    unittest.main()



import json

from inspect import signature

import numpy as np

from AaronTools.atoms import Atom
from AaronTools.comp_output import CompOutput
from AaronTools.component import Component
from AaronTools.spectra import Frequency, HarmonicVibration
from AaronTools.finders import (
    Finder,
    AnyNonTransitionMetal,
    AnyTransitionMetal,
    NotAny,
    get_class,
)
from AaronTools.geometry import Geometry
from AaronTools.substituent import Substituent
from AaronTools.theory import (
    Theory,
    ImplicitSolvent,
    OptimizationJob,
    FrequencyJob,
    ForceJob,
    SinglePointJob,
    BasisSet,
    Basis,
    ECP,
)


class ATEncoder(json.JSONEncoder):
    def default(self, obj):
        """
        Calls appropriate encoding method for supported AaronTools types.
        If type not supported, calls the default `default` method
        """
        if isinstance(obj, Atom):
            return self._encode_atom(obj)
        elif isinstance(obj, Geometry):
            return self._encode_geometry(obj)
        elif isinstance(obj, CompOutput):
            return self._encode_comp_output(obj)
        elif isinstance(obj, Frequency):
            return self._encode_frequency(obj)
        elif isinstance(obj, Theory):
            return self._encode_theory(obj)
        elif isinstance(obj, Finder):
            return self._encode_finder(obj)
        else:
            super().default(obj)

    def _encode_atom(self, obj):
        """
        Encodes the data necessary to re-inialize an equivalent atom.

        Note: constraint information is lost and must be re-initialized by the
        parent geometry through the use of Geometry.parse_comment()
        """
        rv = {"_type": obj.__class__.__name__}
        rv["element"] = obj.element
        rv["coords"] = list(obj.coords)
        rv["flag"] = obj.flag
        rv["name"] = obj.name
        rv["tags"] = list(sorted(obj.tags))
        rv["charge"] = obj.charge
        rv["_rank"] = obj._rank
        return rv

    def _encode_geometry(self, obj):
        """
        Encodes data necessary to re-initialize a geometry object.
        """
        rv = {"_type": obj.__class__.__name__}

        # for Geometry and all child classes
        rv["name"] = obj.name
        rv["atoms"] = obj.atoms
        rv["connectivity"] = []
        for a in obj.atoms:
            rv["connectivity"] += [[obj.atoms.index(b) for b in a.connected]]

        # for Geometry and all child classes but Substituent
        if hasattr(obj, "comment"):
            rv["comment"] = obj.comment

        # for Catalyst child classes
        if isinstance(obj, Geometry) and obj.components:
            # comment
            obj.fix_comment()
            rv["comment"] = obj.comment

        # for Substituent child class
        if hasattr(obj, "conf_num"):
            rv["conf_num"] = obj.conf_num
        if hasattr(obj, "conf_angle"):
            rv["conf_angle"] = obj.conf_angle
        if hasattr(obj, "end"):
            rv["end"] = obj.end

        # for Component child class
        if hasattr(obj, "key_atoms"):
            rv["key_atoms"] = obj.key_atoms
        return rv

    def _encode_comp_output(self, obj):
        rv = {"_type": obj.__class__.__name__}
        rv["geometry"] = obj.geometry
        rv["opts"] = obj.opts
        rv["frequency"] = obj.frequency
        rv["archive"] = obj.archive

        rv["E_ZPVE"] = obj.E_ZPVE
        rv["ZPVE"] = obj.ZPVE
        rv["energy"] = obj.energy
        rv["enthalpy"] = obj.enthalpy
        rv["free_energy"] = obj.free_energy
        rv["grimme_g"] = obj.grimme_g

        rv["mass"] = obj.mass
        rv["charge"] = obj.charge
        rv["multiplicity"] = obj.multiplicity
        rv["temperature"] = obj.temperature

        rv["gradient"] = obj.gradient
        rv["rotational_symmetry_number"] = obj.rotational_symmetry_number
        rv["rotational_temperature"] = obj.rotational_temperature

        rv["error"] = obj.error
        rv["error_msg"] = obj.error_msg
        rv["finished"] = obj.finished

        return rv

    def _encode_frequency(self, obj):
        rv = {"_type": obj.__class__.__name__}
        data = []
        for d in obj.data:
            entry = {}
            for k, v in d.__dict__.items():
                if isinstance(v, np.ndarray):
                    v = v.tolist()
                entry[k] = v
            data += [entry.copy()]
        rv["data"] = data
        return rv

    def _encode_theory(self, obj):
        rv = {"_type": obj.__class__.__name__}
        
        if obj.method:
            rv["method"] = obj.method.name
            rv["semi-empirical"] = obj.method.is_semiempirical
        if obj.grid:
            rv["grid"] = obj.grid.name
        if obj.empirical_dispersion:
            rv["disp"] = obj.empirical_dispersion.name
        if obj.solvent:
            rv["solvent model"] = obj.solvent.solvent_model
            rv["solvent"] = obj.solvent.solvent
        if obj.processors:
            rv["nproc"] = obj.processors
        if obj.memory:
            rv["mem"] = obj.memory
        if obj.kwargs:
            rv["other"] = obj.kwargs

        if obj.job_type:
            rv["jobs"] = {}
            for job in obj.job_type:
                job_type = job.__class__.__name__
                rv["jobs"][job_type] = {}
                for arg in signature(job.__init__).parameters:
                    if arg == "self" or arg == "geometry" or "*" in arg:
                        continue
                    try:
                        rv["jobs"][job_type][arg] = getattr(job, arg)
                    except AttributeError:
                        pass
                
        if obj.basis:
            rv["basis"] = {"name": [], "elements":[], "file":[], "auxiliary":[]}
            if obj.basis.basis:
                for basis in obj.basis.basis:
                    rv["basis"]["name"].append(basis.name)
                    rv["basis"]["elements"].append([])
                    for ele in basis.ele_selection:
                        if isinstance(ele, str):
                            rv["basis"]["elements"][-1].append(ele)
                        elif isinstance(ele, AnyTransitionMetal):
                            rv["basis"]["elements"][-1].append("tm")
                        elif isinstance(ele, AnyNonTransitionMetal):
                            rv["basis"]["elements"][-1].append("!tm")
                            
                    if basis.not_anys:
                        for ele in basis.not_anys:
                            if isinstance(ele, str):
                                rv["basis"]["elements"][-1].append("!%s" % ele)
                            elif isinstance(ele, AnyTransitionMetal):
                                rv["basis"]["elements"][-1].append("!tm")
                            elif isinstance(ele, AnyNonTransitionMetal):
                                rv["basis"]["elements"][-1].append("!!tm")
    
    
                    rv["basis"]["file"].append(basis.user_defined)
                    rv["basis"]["auxiliary"].append(basis.aux_type)

            if obj.basis.ecp:
                rv["ecp"] = {"name": [], "elements":[], "file":[]}
                for basis in obj.basis.ecp:
                    rv["ecp"]["name"].append(basis.name)
                    rv["ecp"]["elements"].append([])
                    for ele in basis.ele_selection:
                        if isinstance(ele, str):
                            rv["ecp"]["elements"][-1].append(ele)
                        elif isinstance(ele, AnyTransitionMetal):
                            rv["ecp"]["elements"][-1].append("tm")
                        elif isinstance(ele, AnyNonTransitionMetal):
                            rv["ecp"]["elements"][-1].append("!tm")

                    if basis.not_anys:
                        for ele in basis.not_anys:
                            if isinstance(ele, str):
                                rv["ecp"]["elements"][-1].append("!%s" % ele)
                            elif isinstance(ele, AnyTransitionMetal):
                                rv["ecp"]["elements"][-1].append("!tm")
                            elif isinstance(ele, AnyNonTransitionMetal):
                                rv["ecp"]["elements"][-1].append("!!tm")
    
                    rv["ecp"]["file"].append(basis.user_defined)

            if obj.kwargs:
                rv["other"] = obj.kwargs

        return rv

    def _encode_finder(self, obj):
        rv = {"_type": "Finder"}
        rv["_spec_type"] = obj.__class__.__name__
        rv["kwargs"] = obj.__dict__
        for kw in rv["kwargs"]:
            if isinstance(rv["kwargs"][kw], np.ndarray):
                rv["kwargs"][kw] = rv["kwargs"][kw].tolist()
        return rv


class ATDecoder(json.JSONDecoder):
    with_progress = False

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(
            self, object_hook=self.object_hook, *args, **kwargs
        )

    def object_hook(self, obj):
        if "_type" not in obj:
            return obj
        if obj["_type"] == "Atom":
            return self._decode_atom(obj)
        if obj["_type"] == "Substituent":
            return self._decode_substituent(obj)
        if obj["_type"] in ["Geometry", "Component"]:
            return self._decode_geometry(obj)
        if obj["_type"] == "Frequency":
            return self._decode_frequency(obj)
        if obj["_type"] == "CompOutput":
            return self._decode_comp_output(obj)
        if obj["_type"] == "Theory":
            return self._decode_theory(obj)
        if obj["_type"] == "Finder":
            return self._decode_finder(obj)

    def _decode_atom(self, obj):
        kwargs = {}
        for key in ["element", "coords", "flag", "name", "tags", "charge"]:
            if key not in obj:
                continue
            kwargs[key] = obj[key]
        rv = Atom(**kwargs)
        rv._rank = obj["_rank"]
        return rv

    def _decode_geometry(self, obj):
        if ATDecoder.with_progress:
            print("Loading structure", obj["name"], " " * 50, end="\r")
        kwargs = {"structure": obj["atoms"]}
        for key in ["name", "comment"]:
            kwargs[key] = obj[key]
        geom = Geometry(**kwargs, refresh_connected=False, refresh_ranks=False)
        for i, connected in enumerate(obj["connectivity"]):
            for c in connected:
                geom.atoms[i].connected.add(geom.atoms[c])

        if obj["_type"] == "Component":
            key_atom_names = [a.name for a in obj["key_atoms"]]
            return Component(geom, key_atoms=key_atom_names)
        else:
            return geom

    def _decode_substituent(self, obj):
        kwargs = {}
        for key in ["name", "end", "conf_num", "conf_angle"]:
            kwargs[key] = obj[key]
        ranks = [a._rank for a in obj["atoms"]]
        obj = self._decode_geometry(obj)
        for a, r in zip(obj.atoms, ranks):
            a._rank = r
        return Substituent(obj, **kwargs)

    def _decode_frequency(self, obj):
        data = []
        for d in obj["data"]:
            kw = {k:v for k, v in d.items()}
            freq = kw.pop("frequency")
            data += [
                HarmonicVibration(freq, **kw)
            ]
        return Frequency(data)

    def _decode_comp_output(self, obj):
        keys = [
            "geometry",
            "opts",
            "frequency",
            "archive",
            "energy",
            "enthalpy",
            "free_energy",
            "grimme_g",
            "gradient",
            "frequency",
            "E_ZPVE",
            "ZPVE",
            "mass",
            "temperature",
            "multiplicity",
            "charge",
            "rotational_temperature",
            "rotational_symmetry_number",
            "error",
            "error_msg",
            "finished",
        ]
        rv = CompOutput()
        for key in keys:
            rv.__dict__[key] = obj[key]
        return rv

    def _decode_theory(self, obj):
        rv = Theory()
        if "method" in obj:
            rv.method = obj["method"]
            if "semi-empirical" in obj:
                rv.method.is_semiempirical = obj["semi-empirical"]
        
        if "grid" in obj:
            rv.grid = obj["grid"]
        
        if "solvent model" in obj and "solvent" in obj:
            rv.solvent = ImplicitSolvent(obj["solvent model"], obj["solvent"])
        
        if "disp" in obj:
            rv.empirical_dispersion = obj["disp"]
        
        if "nproc" in obj:
            rv.processors = obj["nproc"]
        
        if "mem" in obj:
            rv.memory = obj["mem"]

        if "jobs" in obj:
            jobs = []
            for job in obj["jobs"]:
                if job == "OptimizationJob":
                    jobs.append(OptimizationJob(**obj["jobs"][job]))
                elif job == "FrequencyJob":
                    jobs.append(FrequencyJob(**obj["jobs"][job]))
                elif job == "SinglePointJob":
                    jobs.append(SinglePointJob(**obj["jobs"][job]))
                elif job == "ForceJob":
                    jobs.append(ForceJob(**obj["jobs"][job]))
            rv.job_type = jobs
        
        if "basis" in obj or "ecp" in obj:
            rv.basis = BasisSet([], [])
        
        if "basis" in obj:
            for name, aux_type, file, elements in zip(
                    obj["basis"]["name"],
                    obj["basis"]["auxiliary"],
                    obj["basis"]["file"],
                    obj["basis"]["elements"],
            ):
                rv.basis.basis.append(
                    Basis(
                        name,
                        elements=elements,
                        aux_type=aux_type,
                        user_defined=file,
                    )
                ) 

        if "ecp" in obj:
            for name, file, elements in zip(
                    obj["ecp"]["name"],
                    obj["ecp"]["file"],
                    obj["ecp"]["elements"],
            ):
                rv.basis.ecp.append(
                    ECP(
                        name,
                        elements=elements,
                        user_defined=file,
                    )
                )

        if "other" in obj:
            rv.kwargs = obj["other"]
        
        return rv

    def _decode_finder(self, obj):
        specific_type = obj["_spec_type"]
        kwargs = obj["kwargs"]
        cls = get_class(specific_type)
        args = []
        sig = signature(cls.__init__)
        for param in sig.parameters.values():
            if param.name in kwargs and (
                param.kind == param.POSITIONAL_ONLY or
                param.kind == param.POSITIONAL_OR_KEYWORD
            ):
                args.append(kwargs.pop(param.name))
        return cls(*args, **kwargs)



#! /usr/bin/env python3
"""Testing for Substituent class"""
import os
import unittest

import numpy as np
from AaronTools.geometry import Geometry
from AaronTools.substituent import Substituent
from AaronTools.test import TestWithTimer, prefix, rmsd_tol, validate


class TestSubstituent(TestWithTimer):
    COCH3 = Geometry(os.path.join(prefix, "test_files", "COCH3.xyz"))
    NO2 = Geometry(os.path.join(prefix, "test_files", "NO2.xyz"))
    benz_NO2_Cl = Geometry(
        os.path.join(prefix, "test_files", "benzene_1-NO2_4-Cl.xyz")
    )

    def is_COCH3(self, sub):
        ref = TestSubstituent.COCH3
        self.assertEqual(sub.name, "COCH3")
        self.assertEqual(sub.comment, "CF:2,180")
        self.assertEqual(sub.conf_num, 2)
        self.assertEqual(sub.conf_angle, np.deg2rad(180))
        self.assertTrue(validate(sub, ref))
        return

    def is_NO2(self, sub):
        ref = TestSubstituent.NO2
        self.assertEqual(sub.name, "NO2")
        self.assertEqual(sub.comment, "CF:2,120")
        self.assertEqual(sub.conf_num, 2)
        self.assertEqual(sub.conf_angle, np.deg2rad(120))
        self.assertTrue(validate(sub, ref, thresh=1e-5))
        return

    def test_init(self):
        sub = Substituent("COCH3")
        self.is_COCH3(sub)

        sub = Substituent("COCH3", targets=["C", "O", "H"])
        self.is_COCH3(sub)
        return

    def test_copy(self):
        sub = Substituent("COCH3")
        sub = sub.copy()
        self.is_COCH3(sub)
        return

    def test_detect_sub(self):
        mol = TestSubstituent.benz_NO2_Cl

        NO2 = mol.get_fragment("N", "C", as_object=True)
        sub = Substituent(NO2)
        self.is_NO2(sub)

        NO2 = mol.get_fragment("N", "C", copy=True)
        sub = Substituent(NO2)
        self.is_NO2(sub)

    def test_align_to_bond(self):
        mol = TestSubstituent.benz_NO2_Cl
        bond = mol.bond("1", "12")

        sub = Substituent("NO2")
        sub.align_to_bond(bond)

        bond /= np.linalg.norm(bond)
        test_bond = sub.find("N")[0].coords - np.array([0.0, 0.0, 0.0])
        test_bond /= np.linalg.norm(test_bond)
        self.assertTrue(np.linalg.norm(bond - test_bond) < 10 ** -8)


if __name__ == "__main__":
    unittest.main()



import sys
import argparse

import numpy as np

from AaronTools.fileIO import FileReader, read_types
from AaronTools.geometry import Geometry
from AaronTools.symmetry import (
    PointGroup,
    InversionCenter, 
    ProperRotation,
    ImproperRotation,
    MirrorPlane,
)
from AaronTools.utils.utils import perp_vector, glob_files, get_filename

pg_parser = argparse.ArgumentParser(
    description="print point group",
    formatter_class=argparse.RawTextHelpFormatter
)

pg_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file"
)

pg_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    required=False,
    dest="outfile",
    help="output destination \nDefault: stdout"
)

pg_parser.add_argument(
    "-if", "--input-format",
    type=str,
    default=None,
    dest="input_format",
    choices=read_types,
    help="file format of input - xyz is assumed if input is stdin"
)

pg_parser.add_argument(
    "-t", "--tolerance",
    default=0.1,
    type=float,
    dest="tolerance",
    help="tolerance for determining if a symmetry element is valid\n"
    "for the input structure(s)\nDefault: 0.1"
)

pg_parser.add_argument(
    "-a", "--axis-tolerance",
    default=0.01,
    type=float,
    dest="rotation_tolerance",
    help="tolerance for determining if two axes are coincident or orthogonal"
    "\nDefault: 0.01"
)

pg_parser.add_argument(
    "-n", "--max-n",
    default=6,
    type=int,
    dest="max_n",
    help="max. order for proper rotation axes (improper rotations can be 2x this)"
    "\nDefault: 6"
)

pg_parser.add_argument(
    "-e", "--report-error",
    action="store_true",
    default=False,
    dest="report_error",
    help="print all symmetry elements",
)


args = pg_parser.parse_args()

s = ""
for f in glob_files(args.infile, parser=pg_parser):
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format, None), just_geom=True)
        else:
            infile = FileReader(f, just_geom=True)
    else:
        if args.input_format is not None:
            infile = FileReader(("from stdin", args.input_format[0], f), just_geom=True)
        else:
            if len(sys.argv) >= 1:
                infile = FileReader(("from stdin", "xyz", f), just_geom=True)

    geom = Geometry(infile)
    
    out = geom.write(outfile=False)

    pg = PointGroup(
        geom,
        tolerance=args.tolerance,
        rotation_tolerance=args.rotation_tolerance,
        max_rotation=args.max_n,
    )
    if args.report_error:
        print("%s is %s" % (geom.name, pg.name))
        tot_error, max_error, max_ele = pg.total_error(return_max=True)
        print("total error before symmetrizing: %.4f" % tot_error)
        print("max. error before symmetrizing: %.4f (%s)" % (max_error, max_ele))
    pg.idealize_geometry()
    if args.report_error:
        tot_error, max_error, max_ele = pg.total_error(return_max=True)
        print("total error after symmetrizing: %.4f" % tot_error)
        print("max. error after symmetrizing: %.4f (%s)" % (max_error, max_ele))

    out = geom.write(outfile=False)
    
    if not args.outfile:
        print(out)
    else:
        outfile = args.outfile
        if "$INFILE" in outfile:
            outfile = outfile.replace("$INFILE", get_filename(f))
        with open(outfile, "w") as f:
            f.write(out)





#!/usr/bin/env python3

import sys
import argparse
from warnings import warn
from AaronTools.geometry import Geometry
from AaronTools.fileIO import FileReader, read_types
from AaronTools.utils.utils import get_filename, glob_files

def range2int(s):
    """split on "," and turn "-" into a range
    range2int(["1,3-5,7"]) returns [0, 2, 3, 4, 6]
    returns None if input is None"""
    if s is None:
        return s

    if isinstance(s, list):
        range_str = ",".join(s)

    out = []
    c = range_str.split(",")
    for v in c:
        n = v.split("-")
        if len(n) == 2:
            out.extend([i for i in range(int(n[0])-1, int(n[1]))])
        else:
            for i in n:
                out.append(int(i)-1)

    return out

rmsd_parser = argparse.ArgumentParser(
    description="align structure to reference",
    formatter_class=argparse.RawTextHelpFormatter
)

rmsd_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file"
)

rmsd_parser.add_argument(
    "-if", "--input-format",
    type=str,
    default=None,
    choices=read_types,
    dest="input_format",
    help="file format of input - xyz is assumed if input is stdin"
)

rmsd_parser.add_argument(
    "-r", "--reference",
    type=str,
    default=None,
    dest="ref",
    help="reference structure"
)

rmsd_parser.add_argument(
    "-it", "--input-targets",
    type=str,
    default=None,
    required=False,
    dest="in_target",
    metavar="targets",
    help="target atoms on input (1-indexed)\n" +
    "comma (,) and/or hyphen (-) separated list\n" +
    "hyphens denote a range of atoms\n" +
    "commas separate individual atoms or ranges\n" +
    "Default: whole structure"
)

rmsd_parser.add_argument(
    "-rt", "--ref-targets",
    type=str,
    default=None,
    required=False,
    dest="ref_target",
    metavar="targets",
    help="target atoms on reference (1-indexed)"
)

output_options = rmsd_parser.add_argument_group("output options")
output_format = output_options.add_mutually_exclusive_group(required=False)
output_format.add_argument(
    "-v", "--value",
    action="store_true",
    required=False,
    dest="value_only",
    help="print RMSD only"
)

output_format.add_argument(
    "-csv", "--comma-seperated",
    action="store_true",
    required=False,
    dest="csv",
    help="print output in CSV format"
)

output_options.add_argument(
    "-d", "--delimiter",
    type=str,
    default="comma",
    dest="delimiter",
    choices=["comma", "semicolon", "tab", "space"],
    help="CSV delimiter"
)

rmsd_parser.add_argument(
    "-s", "--sort",
    action="store_true",
    required=False,
    dest="sort",
    help="sort atoms"
)

rmsd_parser.add_argument(
    "-n", "--non-hydrogen",
    action="store_true",
    required=False,
    dest="heavy",
    help="ignore hydrogen atoms"
)

output_options.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    required=False,
    dest="outfile",
    help="output destination\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: stdout"
)

args = rmsd_parser.parse_args()

if args.ref is not None:
    ref_geom = Geometry(args.ref)
else:
    rmsd_parser.print_help()
    raise RuntimeError("reference geometry was not specified")

if bool(args.in_target) ^ bool(args.ref_target):
    warn("targets may need to be specified for both input and reference")

if args.csv:
    if args.delimiter == "comma":
        delim = ","
    elif args.delimiter == "space":
        delim = " "
    elif args.delimiter == "semicolon":
        delim = ";"
    elif args.delimiter == "tab":
        delim = "\t"

    header = delim.join(["reference", "geometry", "RMSD"])

    if args.outfile:
        with open(args.outfile, "w") as f:
            f.write(header + "\n")
    else:
        print(header)


for f in glob_files(args.infile, parser=rmsd_parser):
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format, None))
        else:
            infile = FileReader(f)
    else:
        if args.input_format is not None:
            infile = FileReader(("from stdin", args.input_format, f))
        else:
            infile = FileReader(("from stdin", "xyz", f))

    geom = Geometry(infile)

    # align
    rmsd = geom.RMSD(
        ref_geom,
        align=True,
        targets=args.in_target,
        ref_targets=args.ref_target,
        heavy_only=args.heavy,
        sort=args.sort
    )

    geom.comment = "rmsd = %f" % rmsd

    if not args.value_only and not args.csv:
        if args.outfile:
            outfile = args.outfile
            if "$INFILE" in outfile:
                outfile = outfile.replace("$INFILE", get_filename(f))
            geom.write(append=True, outfile=outfile)
        else:
            print(geom.write(outfile=False))

    elif args.value_only:
        if args.outfile:
            outfile = args.outfile
            if "$INFILE" in outfile:
                outfile = outfile.replace("$INFILE", get_filename(f))
            with open(outfile, "a") as f:
                f.write("%f\n" % rmsd)

        else:
            print("%f" % rmsd)

    elif args.csv:
        s = delim.join([ref_geom.name, geom.name, "%f" % rmsd])
        if args.outfile:
            outfile = args.outfile
            if "$INFILE" in outfile:
                outfile = outfile.replace("$INFILE", get_filename(f))
            with open(outfile, "a") as f:
                f.write(s + "\n")
        else:
            print(s)



#!/usr/bin/env python3

import sys
import argparse

from AaronTools.comp_output import CompOutput
from AaronTools.fileIO import FileReader
from AaronTools.utils.utils import glob_files


stat_parser = argparse.ArgumentParser(
    description="prints status of optimization job",
    formatter_class=argparse.RawTextHelpFormatter
)

stat_parser.add_argument(
    "infile",
    metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="input optimization file (i.e. Gaussian output where \"opt\" was specified)"
)

stat_parser.add_argument(
    "-if", "--input-format",
    type=str,
    nargs=1,
    default=None,
    dest="input_format",
    choices=["log", "out", "dat"],
    help="file format of input - required if input is stdin"
)

stat_parser.add_argument(
    "-o", "--output",
    type=str,
    nargs="+",
    default=None,
    required=False,
    dest="outfile",
    help="output destination\nDefault: stdout"
)

args = stat_parser.parse_args()

s = ""

header_vals = [None]

for f in glob_files(args.infile, parser=stat_parser):
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format[0], None), just_geom=False)
        else:
            infile = FileReader(f, just_geom=False)
    else:
        if args.input_format is not None:
            infile = FileReader(("from stdin", args.input_format[0], f), just_geom=False)
        else:
            stat_parser.print_help()
            raise RuntimeError(
                "when no input file is given, stdin is read and a format must be specified"
            )

    co = CompOutput(infile)
    if co.gradient.keys() and (
        not all(
            x in header_vals for x in co.gradient.keys()
        ) or not all(
            x in co.gradient for x in header_vals
        )
    ):
        header_vals = [x for x in sorted(co.gradient.keys())]
        header = "                      Filename    Step  " + "  ".join(
            ["%14s" % crit for crit in header_vals]
        )
        header += "\n"
        s += header

    s += "%30s" % f
    s += "%8s" % co.opt_steps

    if co.gradient.keys():
        for crit in header_vals:
            col = "%.2e/%s" % (
                float(co.gradient[crit]["value"]), "YES" if co.gradient[crit]["converged"] else "NO"
            )
            s += "  %14s" % col

    if (
            "error" in infile.other and
            infile.other["error"] is not None and
            infile.other["error"] != "UNKNOWN"
    ):
        s += "  %s" % infile.other["error_msg"]
    elif not co.gradient.keys():
        s += "  no progress found"
    elif co.finished:
        s += "  finished"
    else:
        s += "  not finished"

    s += "\n"


if not args.outfile:
    print(s.rstrip())
else:
    with open(
            args.outfile,
            "a"
    ) as f:
        f.write(s.rstrip())
